#!/usr/bin/env 
  
__author__ = 'Florian Hase'

#========================================================================

import sys

from utilities import PhoenicsModuleError, PhoenicsVersionError

#========================================================================

try:
    import tensorflow as tf
except ModuleNotFoundError:
    _, error_message, _ = sys.exc_info()
    extension = '\n\tTry installing the tensorflow package or use a different backend instead.'
    PhoenicsModuleError(str(error_message) + extension)

if not tf.__version__ in ['1.4.0', '1.4.1']:
    PhoenicsVersionError('cannot operate with tensorflow version: "%s".\n\tPlease install version 1.4.1' % tf.__version__)

try: 
	import edward as ed
except ModuleNotFoundError:
	_, error_message, _ = sys.exc_info()
	extension = '\n\tTry installing the edward package or use a different backend instead.'
	PhoenicsModuleError(str(error_message) + extension)

if not ed.__version__ in ['1.3.5']:
	PhoenicsVersionError('cannot operate with edward version: "%s".\n\tPlease install version 1.3.5' % ed.__version__)

#========================================================================


import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
tf.logging.set_verbosity(tf.logging.ERROR)

from BayesianNetwork.EdwardInterface.numpy_graph      import NumpyGraph
from BayesianNetwork.EdwardInterface.edward_interface import EdwardNetwork




#!/usr/bin/env python

__author__ = 'Florian Hase'

#=======================================================================

import numpy as np 
from   datetime import datetime

from . import DB_Werkzeug
from gryffin.utilities       import Logger
from gryffin.utilities       import GryffinUnknownSettingsError, GryffinValueError

#=======================================================================

class DatabaseHandler(DB_Werkzeug, Logger):

	DB_ATTRIBUTES = {
			'descriptor_summary': 'pickle',
			'end_time':           'string',
			'received_obs':       'pickle', 
			'runtime':            'string',
			'start_time':         'string',
			'suggested_params':   'pickle',
		}


	def __init__(self, config):
		self.config = config
		DB_Werkzeug.__init__(self, self.config, self.DB_ATTRIBUTES, verbosity = self.config.get('verbosity'))
		Logger.__init__(self, 'DatabaseHandler', self.config.get('verbosity'))

		self.create_database()
		self.create_cache()


	def save(self, db_entry):
		if self.config.get_db('log_runtimes'):
			db_entry['runtime'] = db_entry['end_time'] - db_entry['start_time']
			for attr in ['start_time', 'end_time', 'runtime']:
				db_entry[attr] = str(db_entry[attr])
		else:
			for attr in ['start_time', 'end_time', 'runtime']:
				db_entry[attr] = 'n/a'
		if not self.config.get_db('log_observations'):
			db_entry[received_obs] = []
		self.db_add(db_entry)


	def read_db(self, outfile, verbose):
	
		db_content = self.db_fetch_all()
		if len(db_content) == 0:
			GryffinValueError('no entries found in database')
		out_format = outfile.split('.')[-1]
		if not out_format in ['csv', 'xlsx', 'pkl', 'json']:
			GryffinUnknownSettingsError('did not understand output format "%s".\n\tPlease choose from "csv", "json", "pkl" or "xlsx"' % out_format)

		# sort entries
		if self.config.get_db('log_runtimes'):
			start_times     = [datetime.strptime(entry['start_time'], '%Y-%m-%d %H:%M:%S.%f') for entry in db_content]
			sorting_indices = np.argsort(start_times)
		else:
			sorting_indices = np.arange(len(db_content))

		# create output dictionary
		relevant_keys         = ['start_time', 'end_time', 'runtime']
		if self.config.get('auto_desc_gen'):
			relevant_keys.append('descriptor_summary')
		first_suggested_batch = db_content[0]['suggested_params']
		len_batch             = len(first_suggested_batch)
		param_names           = list(first_suggested_batch[0].keys())
		for sugg_index in range(len_batch):
			for param_name in param_names:
				relevant_keys.append('%s (%d)' % (param_name, sugg_index))
		db_dict = {key: [] for key in relevant_keys}

		for sorting_index in sorting_indices:
			entry = db_content[sorting_index]
			for key in entry.keys():
				if key == 'suggested_params':
					for sugg_index in range(len_batch):
						for param_name in param_names:
							if not param_name in entry[key][sugg_index]:
								GryffinValueError('could not find parameter "%s" in db entry' % param_name)
							sugg_params = np.squeeze(entry[key][sugg_index][param_name])
							db_key      = '%s (%d)' % (param_name, sugg_index)
							db_dict[db_key].append(sugg_params)
				else:
					if not key in relevant_keys: continue
					db_dict[key].append(entry[key])

		# set up writer
		if out_format in ['csv', 'xlsx']:
			from DatabaseHandler.PandasWriters import DB_Writer
		elif out_format in ['json']:
			from DatabaseHandler.JsonWriters   import DB_Writer
			db_dict['config'] = self.config.settings
		elif out_format in ['pkl']:
			from DatabaseHandler.PickleWriters import DB_Writer

		self.db_writer = DB_Writer(self.config)
		self.db_writer.write(db_dict, outfile, out_format)





#!/usr/bin/env python

__author__ = 'Florian Hase'

#========================================================================

import os
import time
import pickle
import subprocess
import numpy as np 

from . import CategoryReshaper
from gryffin.utilities import Logger
from gryffin.utilities import GryffinUnknownSettingsError

from .kernel_evaluations import KernelEvaluator

#========================================================================

class BayesianNetwork(Logger):

	def __init__(self, config, model_details = None):
	
		self.COUNTER = 0
		self.has_sampled = False
		self.config = config
		verbosity = self.config.get('verbosity')
		if 'bayesian_network' in verbosity:
			verbosity = verbosity['bayesian_network']
		Logger.__init__(self, 'BayesianNetwork', verbosity = verbosity)
		self.kernel_contribution = lambda x: (np.sum(x), 1.)
		self.cat_reshaper = CategoryReshaper(self.config)

		# get bnn model detals
		if model_details == None:
			from .model_details import model_details
		self.model_details = model_details

		# set up bnn
		if self.config.get('backend') == 'tensorflow':
			from .tfprob_interface import TfprobNetwork 
			self.network_executable = '{}/bayesian_network/tfprob_interface/tfprob_interface.py'.format(self.config.get('home'))
		else:
			GryffinUnknownSettingsError('did not understand backend: "%s".\n\tChoose from "tensorflow"' % self.config_general.backend)

		# get domain volume
		self.volume     = 1.
		feature_lengths = self.config.feature_lengths 
		feature_ranges  = self.config.feature_ranges
		for feature_index, feature_type in enumerate(self.config.feature_types):
			if feature_type == 'continuous':
				self.volume *= feature_ranges[feature_index]
			elif feature_type == 'categorical':
				self.volume *= feature_lengths[feature_index]
			elif feature_type == 'discrete':
				self.volume *= feature_ranges[feature_index]
			else:
				GryffinUnknownSettingsError('did not understand parameter type: "%s" of variable "%s".\n\t(%s) Please choose from "continuous" or "categorical"' % (feature_type, self.config.feature_names[feature_index], self.template))
		self.inverse_volume = 1 / self.volume

		# compute sampling parameter values
		if self.config.get('sampling_strategies') == 1:
			self.sampling_param_values = np.zeros(1)
		else:
			self.sampling_param_values = np.linspace(-1.0, 1.0, self.config.get('sampling_strategies'))
			self.sampling_param_values = self.sampling_param_values[::-1]
		self.sampling_param_values *= self.inverse_volume



	def sample(self, obs_params, obs_objs, num_epochs = None):

		# package received variables 
		sim_data = {'config': self.config, 'model_details': self.model_details, 'obs_params': obs_params, 'obs_objs': obs_objs}
		sim_file = '%s/sampling_information.pkl' % (self.config.get('scratch_dir'))
		with open(sim_file, 'wb') as content:
			pickle.dump(sim_data, content)
		results_file = '%s/sampling_results.pkl' % (self.config.get('scratch_dir'))

		# submit network sampling
		subprocess.call('python %s %s %s %s' % (self.network_executable, self.config.get('home'), sim_file, results_file), shell = True)

		# pick up 
		with open(results_file, 'rb') as content:
			results_dict = pickle.loads(content.read())
		self.trace_kernels = results_dict['trace_kernels']
		self.obs_objs      = results_dict['obs_objs']

		# set sampling to true
		self.has_sampled = True


	def build_kernels(self, descriptors):
		assert self.has_sampled
		trace_kernels = self.trace_kernels
		obs_objs      = self.obs_objs
	
		# shape of the tensors below: (# samples, # obs, # kernels)		
		burnin, thinning = self.model_details['burnin'], self.model_details['thinning']
		locs       = trace_kernels['locs']
		sqrt_precs = trace_kernels['sqrt_precs']
		probs      = trace_kernels['probs']

		start = time.time()
		try:
			if np.all(np.isfinite(np.array(descriptors, dtype = np.float))):
				probs = self.cat_reshaper.reshape(probs, descriptors)
		except ValueError:
			probs = self.cat_reshaper.reshape(probs, descriptors)
		end = time.time()
		#print('ELAPSED TIME (cat reshaping)', end - start)
		self.log('ELAPSED TIME (cat reshaping) ' + str(end - start), 'INFO')

		# write kernel types
		kernel_type_strings = self.config.kernel_types
		kernel_types = []
		for kernel_type_string in kernel_type_strings:
			if kernel_type_string == 'continuous':
				kernel_types.append(0)
			elif kernel_type_string == 'categorical':
				kernel_types.append(1)
			elif kernel_type_string == 'discrete':
				kernel_types.append(1)
		kernel_types = np.array(kernel_types, dtype = np.int32)
		kernel_sizes = self.config.kernel_sizes.astype(np.int32)

		# get lower prob bound
		if self.config.get('boosted'):
			lower_prob_bound = 1e-1
			for size in self.config.feature_ranges:
				lower_prob_bound *= 1. / size
		else:
			lower_prob_bound = 1e-25


		self.kernel_evaluator = KernelEvaluator(locs, sqrt_precs, probs, kernel_types, kernel_sizes, lower_prob_bound, obs_objs, self.inverse_volume)
#		self.kernel_evaluator = KernelEvaluator(locs, sqrt_precs, probs, kernel_types, kernel_sizes, kernel_indices, obs_objs)


		# check if we can construct a cache
		self.caching = np.sum(kernel_types) == len(kernel_types)
		self.cache   = {}
		

		# kernel contribution should support caching	
		def kernel_contribution(proposed_sample):

			if self.caching:
				sample_string = '-'.join([str(int(element)) for element in proposed_sample])
				if sample_string in self.cache:
					num, inv_den  = self.cache[sample_string]
				else:
					num, inv_den, _ = self.kernel_evaluator.get_kernel(proposed_sample.astype(np.float64))
					self.cache[sample_string] = (num, inv_den)		
			else:
				num, inv_den, _ = self.kernel_evaluator.get_kernel(proposed_sample.astype(np.float64))
			return num, inv_den

		self.kernel_contribution = kernel_contribution













#!/usr/bin/env python 
  
# cython: profile=True

__author__ = 'Florian Hase'

#========================================================================

import  cython 
cimport cython

from cython.parallel import prange

import  numpy as np 
cimport numpy as np 

from libc.math cimport exp, abs, round

#========================================================================

cdef double _gauss(double x, double loc, double sqrt_prec):
	cdef double argument, result
	argument = 0.5 * ( sqrt_prec * (x - loc) )**2
	if argument > 200.:
		result = 0.
	else:
		result   = exp( - argument) * sqrt_prec * 0.3989422804014327	# the number is 1. / np.sqrt(2 * np.pi)
	return result

#========================================================================

cdef class KernelEvaluator:

	cdef int    num_samples, num_obs, num_kernels, num_cats
	cdef double lower_prob_bound, inv_vol
	
	cdef np.ndarray np_locs, np_sqrt_precs, np_cat_probs
	cdef np.ndarray np_kernel_types, np_kernel_sizes
	cdef np.ndarray np_objs
	cdef np.ndarray np_probs

	var_dict = {}

	def __init__(self, locs, sqrt_precs, cat_probs, kernel_types, kernel_sizes, lower_prob_bound, objs, inv_vol):
		
		self.np_locs          = locs
		self.np_sqrt_precs    = sqrt_precs
		self.np_cat_probs     = cat_probs
		self.np_kernel_types  = kernel_types
		self.np_kernel_sizes  = kernel_sizes
		self.np_objs          = objs

		self.num_samples      = locs.shape[0]
		self.num_obs          = locs.shape[1]
		self.num_kernels      = locs.shape[2]
		self.lower_prob_bound = lower_prob_bound
		self.inv_vol          = inv_vol

		self.np_probs = np.zeros(self.num_obs, dtype = np.float64)


	@cython.boundscheck(False)
	@cython.cdivision(True)
	cdef double [:] _probs(self, double [:] sample):

		cdef int    sample_index, obs_index, feature_index, kernel_index
		cdef int    num_indices
		cdef int    num_continuous, num_categorical
		cdef double total_prob, prec_prod, exp_arg_sum

		cdef double [:, :, :] locs       = self.np_locs
		cdef double [:, :, :] sqrt_precs = self.np_sqrt_precs 
		cdef double [:, :, :] cat_probs  = self.np_cat_probs

		cdef int [:] kernel_types = self.np_kernel_types
		cdef int [:] kernel_sizes = self.np_kernel_sizes

		cdef double inv_sqrt_two_pi = 0.3989422804014327

		cdef double [:] probs = self.np_probs
		for obs_index in range(self.num_obs):
			probs[obs_index] = 0.

		cdef double cat_prob
		cdef double obs_probs

		# get number of continuous variables
		num_continuous = 0
		while kernel_index < self.num_kernels:
			num_continuous += 1
			kernel_index   += kernel_sizes[kernel_index]


#		print(self.num_obs, self.num_samples, self.num_kernels)

		for obs_index in range(self.num_obs):
			obs_probs = 0.

			for sample_index in range(self.num_samples):
				total_prob     = 1.
				prec_prod      = 1.
				exp_arg_sum    = 0.
				feature_index, kernel_index = 0, 0

				while kernel_index < self.num_kernels:

					if kernel_types[kernel_index] == 0:

						prec_prod      = prec_prod * sqrt_precs[sample_index, obs_index, kernel_index]
						exp_arg_sum    = exp_arg_sum + (sqrt_precs[sample_index, obs_index, kernel_index] * (sample[feature_index] - locs[sample_index, obs_index, kernel_index]))**2

					elif kernel_types[kernel_index] == 1:
						total_prob *= cat_probs[sample_index, obs_index, kernel_index + <int>round(sample[feature_index])]

					kernel_index  += kernel_sizes[kernel_index]
					feature_index += 1

				obs_probs += total_prob * prec_prod * exp( - 0.5 * exp_arg_sum) #* inv_sqrt_two_pi**num_continuous

				if sample_index == 100:
					if 0.01 * obs_probs * inv_sqrt_two_pi**num_continuous < self.lower_prob_bound:
						probs[obs_index] = 0.01 * obs_probs
						break

			else:
				probs[obs_index] = obs_probs * inv_sqrt_two_pi**num_continuous / self.num_samples
		return probs
		return probs





#	@cython.boundscheck(False)	
#	@cython.cdivision(True)
	cdef double [:] _OLD_probs(self, double [:] sample):

		cdef int sample_index, obs_index, feature_index, kernel_index
		cdef int num_indices
		cdef double total_prob

		cdef double [:, :, :] locs       = self.np_locs
		cdef double [:, :, :] sqrt_precs = self.np_sqrt_precs 
		cdef double [:, :, :] cat_probs  = self.np_cat_probs
	
		cdef int [:] kernel_types     = self.np_kernel_types
		cdef int [:] kernel_sizes     = self.np_kernel_sizes 

		cdef double [:] probs = self.np_probs
		for obs_index in range(self.num_obs):
			probs[obs_index] = 0.

		cdef double cat_prob
		cdef double obs_probs
	
#		for sample_index in range(self.num_samples):
		for obs_index in range(self.num_obs):		
			
			obs_probs = 0.

#			for obs_index in range(self.num_obs):
			for sample_index in range(self.num_samples):
	
				total_prob    = 1.
				feature_index = 0
				kernel_index  = 0

				while kernel_index < self.num_kernels:
#				for feature_index in range(num_indices):	

					if kernel_types[kernel_index] == 0:
						total_prob *= _gauss(sample[feature_index], locs[sample_index, obs_index, kernel_index], sqrt_precs[sample_index, obs_index, kernel_index])

					elif kernel_types[kernel_index] == 1:

						cat_prob = cat_probs[sample_index, obs_index, kernel_index + <int>round(sample[feature_index])]
#						if cat_prob < lower_prob_bounds[feature_index]:
#						if cat_prob < 1e-2:
#							break

						total_prob *= cat_prob

					kernel_index  += kernel_sizes[kernel_index]
					feature_index += 1

#				else:
				obs_probs += total_prob

				if sample_index == 100:
					if 0.01 * obs_probs < self.lower_prob_bound:
						probs[obs_index] = 0.01 * obs_probs
						break

#			probs[obs_index] = obs_probs / self.num_samples
			else:
				probs[obs_index] = obs_probs / self.num_samples
#			print(np.asarray(probs), np.asarray(sample), np.asarray(kernel_types))
		return probs


#	@cython.boundscheck(False)
	cpdef get_kernel(self, np.ndarray sample):

		cdef int obs_index
		cdef double temp_0, temp_1
		cdef double inv_den
		
		cdef double [:] sample_memview = sample
		probs_sample = self._probs(sample_memview)

		# construct numerator and denominator of acquisition
		cdef double num = 0.
		cdef double den = 0.
		cdef double [:] objs = self.np_objs 

		for obs_index in range(self.num_obs):
			temp_0 = objs[obs_index]
			temp_1 = probs_sample[obs_index]
			num += temp_0 * temp_1
			den += temp_1

		inv_den = 1. / (self.inv_vol + den)

#		print('-->', num, den)
#		print('OBJETIVES', self.np_objs)

		return num, inv_den, probs_sample
	








#!/usr/bin/env python

from .abstract_optimizer          import AbstractOptimizer

from .adam_optimizer              import AdamOptimizer
from .naive_categorical_optimizer import NaiveCategoricalOptimizer
from .naive_discrete_optimizer    import NaiveDiscreteOptimizer




#!/usr/bin/env python
  
__author__ = 'Florian Hase'

#=======================================================================

import sys

from utilities import PhoenicsModuleError

#=======================================================================

try:
	import pandas
except ModuleNotFoundError:
	_, error_message, _ = sys.exc_info()
	extension = '\n\tTry installing the pandas package of use a different database output format (e.g. pickle).'
	PhoenicsModuleError(str(error_message) + extension)

try:
	import openpyxl
except ModuleNotFoundError:
	_, error_message, _ = sys.exc_info()
	extension = '\n\tWriting to xlsx requires the openpyxl package. Please install the package or choose a different output format (e.g. csv or pickle)'
	PhoenicsModuleError(str(error_message) + extension)

#=======================================================================

from DatabaseHandler.PandasWriters.db_writer import DB_Writer




#!/usr/bin/env python

__author__ = 'Florian Hase'

#=======================================================================
# DEFAULT GENERAL CONFIGURATIONS

default_general_configurations = {
	'auto_desc_gen':          False,
	'backend':               'tensorflow',
	'batches':                1,
	'boosted':                True,
	'parallel':               True,
	'random_seed':            100691,
	'sampler':               'uniform',
	'sampling_strategies':    2,
	'scratch_dir':           './.scratch',
	'softness':               0.001,
	'continuous_optimizer':  'adam',
	'categorical_optimizer': 'naive',
	'discrete_optimizer':    'naive',
	'verbosity': {
			'default':          2,
			'bayesian_network': 3,
			'random_sampler':   2,
		}
}

#=======================================================================
# DEFAULT DATABASE CONFIGURATIONS

default_database_configurations = {
	'format':           'sqlite',
	'path':             './SearchProgress',
	'log_observations':  True,
	'log_runtimes':      True,
}

#=======================================================================

default_configuration = {
	'general': {
			key: default_general_configurations[key] for key in default_general_configurations.keys()
		},
    'database': {
			key: default_database_configurations[key] for key in default_database_configurations.keys()
		},
	'parameters': [
			{'name': 'param_0', 'type': 'continuous', 'low':  0.0, 'high': 10.0, 'size': 2},
			{'name': 'param_1', 'type': 'continuous', 'low': -1.0, 'high':  1.0, 'size': 1},
		],
	'objectives': [ 
			{'name': 'obj_0', 'goal': 'minimize', 'hierarchy': 0, 'tolerance': 0.2},
			{'name': 'obj_1', 'goal': 'maximize', 'hierarchy': 1, 'tolerance': 0.2},
		]
}

#=======================================================================





#!/usr/bin/env 

__author__ = 'Florian Hase'

#========================================================================

import json 

#========================================================================

class ParserJSON(object):

	def __init__(self, json_file = None):
		self.json_file = json_file

	def parse(self, json_file = None):
		# update json file
		if not json_file is None:
			self.json_file = json_file

		# parse configuration
		if not self.json_file is None:
			with open(self.json_file) as content:
				self.parsed_json = json.loads(content.read())
		else:
			from utilities.defaults import default_configuration
			self.parsed_json = default_configuration
		return self.parsed_json




#!/usr/bin/env python
  
__author__ = 'Florian Hase'

#========================================================================

import warnings
warnings.filterwarnings('ignore')

import pickle
import numpy as np 
import tensorflow as tf
import edward     as ed
from multiprocessing import Process, Manager

import os, sys
sys.path.append(os.getcwd())
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from numpy_graph import NumpyGraph

#========================================================================

class EdwardNetwork(object):

	def __init__(self, config, model_details):
		self.config        = config
#		Logger.__init__(self, 'EdwardNetwork', self.config.get('verbosity'))
		self.numpy_graph   = NumpyGraph(self.config, model_details)

		self.model_details = model_details
		for key, value in self.model_details.items():
			setattr(self, '_%s' % str(key), value)

		self.feature_size    = len(self.config.kernel_names)
		self.bnn_output_size = len(self.config.kernel_names)
		self.target_size     = len(self.config.kernel_names)

		self.graph = tf.Graph()
		with self.graph.as_default():
			self.sess = tf.InteractiveSession()


	def declare_training_data(self, obs_params, obs_objs):

		assert len(obs_params) == len(obs_objs)
		# categorical features are encoded as single numbers, bnn output is one-hot encoded
		self.num_obs = len(obs_params)
		self.obs_objs = obs_objs

		
		# initialize training features and targets
		self.features = np.zeros((self.num_obs, self.feature_size))
		self.targets  = np.zeros((self.num_obs, self.target_size)) 

		# construct training features
		feature_begin = 0
		feature_sizes = self.config.feature_sizes
		for feature_index, feature_type in enumerate(self.config.feature_types):
			feature_size = feature_sizes[feature_index]
			if feature_type == 'categorical':
				for obs_param_index, obs_param in enumerate(obs_params):
					self.features[obs_param_index, feature_begin + int(obs_param[feature_index])] += 1
			elif feature_type == 'discrete':
				for obs_param_index, obs_param in enumerate(obs_params):
					self.features[obs_param_index, feature_begin + int(obs_param[feature_index])] += 1
			elif feature_type == 'continuous':
				self.features[:, feature_begin] = obs_params[:, feature_index]		
			else:
				raise NotImplementedError
			feature_begin += feature_size
		self.targets = self.features.copy()

		# rescale features
		lower_rescalings = np.empty(self.feature_size)
		upper_rescalings = np.empty(self.feature_size)
		kernel_uppers, kernel_lowers = self.config.kernel_uppers, self.config.kernel_lowers
		for kernel_index, kernel_name in enumerate(self.config.kernel_names):
			low = kernel_lowers[kernel_index]
			up  = kernel_uppers[kernel_index]
			lower_rescalings[kernel_index] = low #- 0.1 * (up - low)
			upper_rescalings[kernel_index] = up  #+ 0.1 * (up - low)

		self.lower_rescalings = lower_rescalings
		self.upper_rescalings = upper_rescalings

		self.rescaled_features = (self.features - self.lower_rescalings) / (self.upper_rescalings - self.lower_rescalings)
		self.rescaled_targets  = (self.targets  - self.lower_rescalings) / (self.upper_rescalings - self.lower_rescalings)

#		self.features = (self.features - lower_rescalings) / (upper_rescalings - lower_rescalings)
#		self.targets  = (self.targets  - lower_rescalings) / (upper_rescalings - lower_rescalings)

#		print(lower_rescalings)
#		print(upper_rescalings)
#		print(self.features)
#		print(self.rescaled_features)
#		quit()

		self.numpy_graph.declare_training_data(self.rescaled_features)


	def construct_model(self):

		with self.graph.as_default():

			self.sess.close()
			self.sess = tf.InteractiveSession()
			self.sess.as_default()

#			self.x = tf.convert_to_tensor(self.rescaled_features,  dtype = tf.float32)
#			self.y = tf.convert_to_tensor(self.rescaled_targets,   dtype = tf.float32)

			self.x = tf.convert_to_tensor(self.rescaled_features,  dtype = tf.float32)
			self.y = tf.convert_to_tensor(self.targets,   dtype = tf.float32)

			# construct preciness
			self.tau_rescaling = np.zeros((self.num_obs, self.bnn_output_size))
			kernel_ranges      = self.config.kernel_ranges
			for obs_index in range(self.num_obs):
				self.tau_rescaling[obs_index] += kernel_ranges
			self.tau_rescaling = self.tau_rescaling**2
	
			# construct weight and bias shapes
#			activations = [tf.nn.tanh]
			activations = [lambda x: x]
			weight_shapes, bias_shapes = [[self.feature_size, self.bnn_output_size]], [[self.bnn_output_size]]
#			for _ in range(1, self._num_layers - 1):
#				activations.append(tf.nn.tanh)
#				weight_shapes.append([self._hidden_shape, self._hidden_shape])
#				bias_shapes.append([self._hidden_shape])
#			activations.append(lambda x: x)
#			weight_shapes.append([self._hidden_shape, self.bnn_output_size])
#			bias_shapes.append([self.bnn_output_size])

			# construct prior
			self.prior_layer_outputs = [self.x]
			self.priors = {}
			for layer_index in range(self._num_layers):
				weight_shape, bias_shape = weight_shapes[layer_index], bias_shapes[layer_index]
				activation = activations[layer_index]
	
				weight = ed.models.Normal(loc = tf.zeros(weight_shape) + self._weight_loc, scale = tf.zeros(weight_shape) + self._weight_scale)
				bias   = ed.models.Normal(loc = tf.zeros(bias_shape)   + self._bias_loc,   scale = tf.zeros(bias_shape)   + self._bias_scale)
				self.priors['weight_%d' % layer_index] = weight
				self.priors['bias_%d'   % layer_index] = bias
	
				prior_layer_output = activation(tf.matmul(self.prior_layer_outputs[-1], weight) + bias)
				self.prior_layer_outputs.append(prior_layer_output)
					
			self.prior_bnn_output = self.prior_layer_outputs[-1]
			self.prior_tau_normed = ed.models.Gamma(self.num_obs + tf.zeros((self.num_obs, self.bnn_output_size)), tf.ones((self.num_obs, self.bnn_output_size)))
			self.prior_tau        = self.prior_tau_normed / self.tau_rescaling
			self.prior_scale	  = ed.models.Deterministic(1. / tf.sqrt(self.prior_tau))
	
			# construct posterior
			self.post_layer_outputs = [self.x]
			self.posteriors = {}
			for layer_index in range(self._num_layers):
				weight_shape, bias_shape = weight_shapes[layer_index], bias_shapes[layer_index]
				activation = activations[layer_index]
	
#				weight = ed.models.Normal(loc = tf.Variable(tf.zeros(weight_shape)), scale = tf.nn.softplus(tf.Variable(tf.zeros(weight_shape))))
#				bias   = ed.models.Normal(loc = tf.Variable(tf.zeros(bias_shape)),   scale = tf.nn.softplus(tf.Variable(tf.zeros(bias_shape))))

				weight = ed.models.Normal(loc = tf.Variable(tf.random_normal(weight_shape)), scale = tf.nn.softplus(tf.Variable(tf.zeros(weight_shape))))
				bias   = ed.models.Normal(loc = tf.Variable(tf.random_normal(bias_shape)),   scale = tf.nn.softplus(tf.Variable(tf.zeros(bias_shape))))

				self.posteriors['weight_%d' % layer_index] = weight
				self.posteriors['bias_%d'   % layer_index] = bias
	
				post_layer_output = activation(tf.matmul(self.post_layer_outputs[-1], weight) + bias)
				self.post_layer_outputs.append(post_layer_output)
	
			self.post_bnn_output = self.post_layer_outputs[-1]				
			self.post_tau_normed = ed.models.Gamma(self.num_obs + tf.Variable(tf.zeros((self.num_obs, self.bnn_output_size))), tf.nn.softplus(tf.Variable(tf.ones((self.num_obs, self.bnn_output_size)))))
			self.post_tau        = self.post_tau_normed / self.tau_rescaling
			self.post_sqrt_tau   = tf.sqrt(self.post_tau)
			self.post_scale	     = ed.models.Deterministic(1. / self.post_sqrt_tau)

			# map bnn output to prediction
			post_kernels = {}
			targets_dict = {}
	
			target_element_index = 0
			kernel_element_index = 0
	
			while kernel_element_index < len(self.config.kernel_names):
						
				kernel_type = self.config.kernel_types[kernel_element_index]
				kernel_size = self.config.kernel_sizes[kernel_element_index]
	
				feature_begin, feature_end = target_element_index, target_element_index + 1 
				kernel_begin, kernel_end   = kernel_element_index, kernel_element_index + kernel_size
	
				prior_relevant = self.prior_bnn_output[:, kernel_begin : kernel_end]
				post_relevant  = self.post_bnn_output[:,  kernel_begin : kernel_end]
	
#				print(kernel_begin, kernel_end, target_element_index)

				if kernel_type == 'continuous': 
						
					target  = self.y[:, kernel_begin : kernel_end]
					lowers, uppers = self.config.kernel_lowers[kernel_begin : kernel_end], self.config.kernel_uppers[kernel_begin : kernel_end]
	
					prior_support = (uppers - lowers) * (1.2 * tf.nn.sigmoid(prior_relevant) - 0.1) + lowers
					post_support  = (uppers - lowers) * (1.2 * tf.nn.sigmoid(post_relevant) - 0.1)  + lowers
	
					prior_predict = ed.models.Normal(prior_support, self.prior_scale[:, kernel_begin : kernel_end])				
					post_predict  = ed.models.Normal(post_support,  self.post_scale[:,  kernel_begin : kernel_end])
	
					targets_dict[prior_predict] = target
					post_kernels['param_%d' % target_element_index] = {
						'loc':       ed.models.Deterministic(post_support),
						'sqrt_prec': ed.models.Deterministic(self.post_sqrt_tau[:, kernel_begin : kernel_end]),
						'scale':     ed.models.Deterministic(self.post_scale[:, kernel_begin : kernel_end])}

				elif kernel_type == 'discrete':
	
					target = tf.cast(self.y[:, kernel_begin : kernel_end], tf.int32)
	
					prior_temperature = 0.5 + 1.0 / self.num_obs
#					prior_temperature = 0.4
					post_temperature  = prior_temperature
	
					prior_support = prior_relevant
					post_support  = post_relevant
	
					prior_predict_relaxed = ed.models.RelaxedOneHotCategorical(prior_temperature, prior_support)
					prior_predict         = ed.models.OneHotCategorical(probs = prior_predict_relaxed)
	
					post_predict_relaxed  = ed.models.RelaxedOneHotCategorical(post_temperature, post_support)
					post_predict          = ed.models.OneHotCategorical(probs = post_predict_relaxed)
	
					targets_dict[prior_predict] = target
					post_kernels['param_%d' % target_element_index] = {'probs': post_predict_relaxed}


					'''
						Temperature annealing schedule:
							- temperature of 100   yields 1e-2 deviation from uniform
							- temperature of  10   yields 1e-1 deviation from uniform
							- temperature of   1   yields *almost* perfect agreement with expectation
							- temperature of   0.1 yields perfect agreement with expectation
					'''



				elif kernel_type == 'categorical':
	
					target = tf.cast(self.y[:, kernel_begin : kernel_end], tf.int32)
	
					prior_temperature = 0.5 + 1.0 / self.num_obs
#					prior_temperature = 0.4
					post_temperature  = prior_temperature
	
					prior_support = prior_relevant
					post_support  = post_relevant
	
					prior_predict_relaxed = ed.models.RelaxedOneHotCategorical(prior_temperature, prior_support)
					prior_predict         = ed.models.OneHotCategorical(probs = prior_predict_relaxed)
	
					post_predict_relaxed  = ed.models.RelaxedOneHotCategorical(post_temperature, post_support)
					post_predict          = ed.models.OneHotCategorical(probs = post_predict_relaxed)
	
					targets_dict[prior_predict] = target
					post_kernels['param_%d' % target_element_index] = {'probs': post_predict_relaxed}


					'''
						Temperature annealing schedule:
							- temperature of 100   yields 1e-2 deviation from uniform
							- temperature of  10   yields 1e-1 deviation from uniform
							- temperature of   1   yields *almost* perfect agreement with expectation
							- temperature of   0.1 yields perfect agreement with expectation
					'''


				else: 
					PhoenicsUnknownSettingsError('did not understand parameter type: "%s".\n\tPlease choose from "continuous" or "categorical' % param_type)

				target_element_index += 1
				kernel_element_index += kernel_size

			self.post_kernels = post_kernels
			self.targets_dict = targets_dict
				

			# inference
			inference_dict = {}
			for key, prior_var in self.priors.items():
				inference_dict[prior_var] = self.posteriors[key]

			self.inference = ed.KLqp(inference_dict, data = self.targets_dict)
			self.optimizer = tf.train.AdamOptimizer(self._learning_rate)
			self.inference.initialize(optimizer = self.optimizer, n_iter = self._num_epochs)

			tf.global_variables_initializer().run()



	def sample_thread(self, kernel_value, kernel_name, return_dict):

		with self.graph.as_default():
			print('starting sampling', kernel_name)
			samples = [kernel_value.sample().eval() for _ in range(self._num_draws)]
			return_dict[kernel_name] = np.array(samples)
			print('completed sampling')



	def sample(self, num_epochs = None, num_draws = None):
		if num_epochs is None: num_epochs = self._num_epochs
		if num_draws  is None: num_draws  = self._num_draws
		
		with self.graph.as_default():

			# run inference
			import time 
			start = time.time()
			for _ in range(num_epochs):
				info_dict = self.inference.update()
				self.inference.print_progress(info_dict)
			self.inference.finalize()
			end = time.time()
			print('[TIME]:  ', end - start, '  (inference)')

			# sample posterior
			self.trace = {}
			posterior_samples = {}
			for key, kernel_parent in self.posteriors.items():
				parent_samples = kernel_parent.sample(self._num_draws).eval()
				posterior_samples[key] = parent_samples
	

			import time 
			start        = time.time()	
			post_kernels = self.numpy_graph.compute_kernels(posterior_samples)
			end          = time.time()
			print('[TIME]:  ', end - start, '  (sampling from posterior)')


			for key in post_kernels.keys():
				self.trace[key] = {}
				kernel_dict = post_kernels[key]
				for kernel_name, kernel_values in kernel_dict.items():
					self.trace[key][kernel_name] = kernel_values


	def get_kernels(self):
		trace_kernels = {'locs': [], 'sqrt_precs': [], 'probs': []}				
		for param_index in range(len(self.config.param_names)):
			post_kernel = self.trace['param_%d' % param_index]

			if 'loc' in post_kernel and 'sqrt_prec' in post_kernel:
				trace_kernels['locs'].append(post_kernel['loc'].astype(np.float64))
				trace_kernels['sqrt_precs'].append(post_kernel['sqrt_prec'].astype(np.float64))
#				trace_kernels['scale'].append(post_kernel['scale'].astype(np.float64))
				trace_kernels['probs'].append(np.zeros(post_kernel['loc'].shape, dtype = np.float64))

			elif 'probs' in post_kernel:
				trace_kernels['locs'].append(np.zeros(post_kernel['probs'].shape, dtype = np.float64))
				trace_kernels['sqrt_precs'].append(np.zeros(post_kernel['probs'].shape, dtype = np.float64))
#				trace_kernels['scale'].append(np.zeros(post_kernel['probs'].shape, dtype = np.float64))
				trace_kernels['probs'].append(post_kernel['probs'].astype(np.float64))
			else:
				raise NotImplementedError

		for key, kernel in trace_kernels.items():
			trace_kernels[key] = np.concatenate(kernel, axis = 2)

		return trace_kernels, self.obs_objs


#========================================================================

if __name__ == '__main__':

	import sys
	home_path    = sys.argv[1]
	sim_file     = sys.argv[2]
	results_file = sys.argv[3]

	sys.path.append(home_path)	

	with open(sim_file, 'rb') as content:
		sim_data = pickle.load(content)

	edward_network = EdwardNetwork(sim_data['config'], sim_data['model_details'])
	edward_network.declare_training_data(sim_data['obs_params'], sim_data['obs_objs'])
	edward_network.construct_model()
	edward_network.sample()

	trace_kernels, obs_objs = edward_network.get_kernels()
	results = {'trace_kernels': trace_kernels, 'obs_objs': obs_objs}
	with open(results_file, 'wb') as content:
		pickle.dump(results, content)





















#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

import sys
import traceback

#========================================================================

class AbstractError(object):

	def __init__(self, message):
		self.__call__(message)

	def __call__(self, message):
		error_traceback = traceback.format_exc()
		error_traceback = '\n'.join(error_traceback.split('\n')[:-2]) + '\n\n'
		error_type      = '\x1b[0;31m%s: %s\x1b[0m\n' % (self.name, message)

		if 'SystemExit' in error_traceback:	
			return None
		
		sys.stderr.write(error_traceback)
		sys.stderr.write(error_type)
		sys.exit()
	

class GryffinModuleError(AbstractError):
	name = 'GryffinModuleError'

class GryffinNotFoundError(AbstractError):
	name = 'GryffinNotFoundError'

class GryffinParseError(AbstractError):
	name = 'GryffinParseError'

class GryffinUnknownSettingsError(AbstractError):
	name = 'GryffinUnknownSettingsError'

class GryffinValueError(AbstractError):
	name = 'GryffinValueError'

class GryffinVersionError(AbstractError):
	name = 'GryffinVersionError'




#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

import os, sys
import numpy as np 
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from .acquisition           import Acquisition
from .bayesian_network      import BayesianNetwork
from .descriptor_generator  import DescriptorGenerator
from .observation_processor import ObservationProcessor
from .random_sampler        import RandomSampler
from .sample_selector       import SampleSelector
from .utilities             import ConfigParser, Logger, GryffinNotFoundError

#========================================================================

class Gryffin(Logger):

	def __init__(self, config_file = None, config_dict = None):
	
		Logger.__init__(self, 'Gryffin', verbosity = 0)

		# parse configuration
		self.config = ConfigParser(config_file, config_dict)
		self.config.parse()
		self.config.set_home(os.path.dirname(os.path.abspath(__file__)))
	
		np.random.seed(self.config.get('random_seed'))	
		self.update_verbosity(self.config.get('verbosity'))
		self.create_folders()

		self.random_sampler       = RandomSampler(self.config.general, self.config.parameters)
		self.obs_processor        = ObservationProcessor(self.config)
		self.descriptor_generator = DescriptorGenerator(self.config)
		self.bayesian_network     = BayesianNetwork(self.config)
		self.acquisition          = Acquisition(self.config)
		self.sample_selector      = SampleSelector(self.config)

		self.iter_counter = 0


	def create_folders(self):

		if not os.path.isdir(self.config.get('scratch_dir')):
			try:
				os.mkdir(self.config.get('scratch_dir'))
			except FileNotFoundError:
				GryffinNotFoundError('Could not create scratch directory: %s' % self.config.get('scratch_dir'))

		if self.config.get_db('has_db') and not os.path.isdir(self.config.get_db('path')):
			try:
				os.mkdir(self.config.get_db('path'))
			except FileNotFoundError:
				GryffinNotFoundError('Could not create database directory: %s' % self.config.get_db('path'))		

		if self.config.get_db('has_db'):
			from .database_handler import DatabaseHandler
			self.db_handler = DatabaseHandler(self.config)



	def recommend(self, observations = None, as_array = False):
		
		from datetime import datetime
		start_time = datetime.now()

		if observations is None:
			# no observations, need to fall back to random sampling
			samples = self.random_sampler.draw(num = self.config.get('batches') * self.config.get('sampling_strategies'))
			if self.config.process_constrained:
				dominant_features = self.config.feature_process_constrained
				samples[:, dominant_features] = samples[0, dominant_features]

		elif len(observations) == 0:
			self.log('Could not find any observations, falling back to random sampling', 'WARNING')
			samples = self.random_sampler.draw(num = self.config.get('batches') * self.config.get('sampling_strategies'))
			if self.config.process_constrained:
				dominant_features = self.config.feature_process_constrained
				samples[:, dominant_features] = samples[0, dominant_features]

		else:
			obs_params, obs_objs = self.obs_processor.process(observations)	

			# run descriptor generation
			if self.config.get('auto_desc_gen') and len(obs_params) > 2:
				self.descriptor_generator.generate(obs_params, obs_objs)
			
			self.bayesian_network.sample(obs_params, obs_objs)

			# extract descriptors and build kernels
			descriptors = self.descriptor_generator.get_descriptors()

			self.bayesian_network.build_kernels(descriptors)
			sampling_param_values   = self.bayesian_network.sampling_param_values
			dominant_strategy_index = self.iter_counter % len(sampling_param_values)
			dominant_strategy_value = np.array([sampling_param_values[dominant_strategy_index]])

			# prepare sample generation / selection
			best_params         = obs_params[np.argmin(obs_objs)]
			kernel_contribution = self.bayesian_network.kernel_contribution

			# if there are process constraining parameters, run those first
			if self.config.process_constrained:
				proposed_samples     = self.acquisition.propose(best_params, kernel_contribution, dominant_strategy_value)
				constraining_samples = self.sample_selector.select(self.config.get('batches'), proposed_samples, kernel_contribution, dominant_strategy_value, obs_params)
			else:
				constraining_samples = None

			# then select the remaining proposals
			proposed_samples = self.acquisition.propose(
					best_params, kernel_contribution, sampling_param_values, 
					dominant_samples  = constraining_samples,
					dominant_strategy = dominant_strategy_index,
				)

			samples = self.sample_selector.select(
					self.config.get('batches'), proposed_samples, kernel_contribution, sampling_param_values, obs_params
				)


		end_time = datetime.now()
		self.log('[TIME]:  ' + str(end_time - start_time) + ',   (overall)', 'INFO')
		#print('[TIME]:  ', end_time - start_time, '  (overall)')
		#print('***********************************************')


		if as_array:
			# return as is
			return_samples = samples
		else:
			# convert to list of dictionaries
			param_names   = self.config.param_names
			param_options = self.config.param_options
			param_types   = self.config.param_types
			sample_dicts  = []
			for sample in samples:
				sample_dict  = {}
				lower, upper = 0, self.config.param_sizes[0]
				for param_index, param_name in enumerate(param_names):
					param_type = param_types[param_index]

					if param_type == 'continuous':
						sample_dict[param_name] = sample[lower:upper]

					elif param_type == 'categorical':
						options = param_options[param_index]
						parsed_options = [options[int(element)] for element in sample[lower:upper]]
						sample_dict[param_name] = parsed_options

					elif param_type == 'discrete':
						options = param_options[param_index]
						parsed_options = [options[int(element)] for element in sample[lower:upper]]
						sample_dict[param_name] = parsed_options

					if param_index == len(self.config.param_names) - 1:
						break
					lower  = upper
					upper += self.config.param_sizes[param_index + 1]
				sample_dicts.append(sample_dict)
			return_samples = sample_dicts


		if self.config.get_db('has_db'):
			db_entry = {'start_time': start_time, 'end_time': end_time, 
						'received_obs': observations, 'suggested_params': return_samples}
			if self.config.get('auto_desc_gen'):
				# get summary of learned descriptors
				descriptor_summary = self.descriptor_generator.get_summary()
				db_entry['descriptor_summary'] = descriptor_summary
			self.db_handler.save(db_entry)

		self.iter_counter += 1
		return return_samples



	def read_db(self, outfile = 'database.csv', verbose = True):
		self.db_handler.read_db(outfile, verbose)


#========================================================================

if __name__ == '__main__':

	observations = [
			{'param_0': [-1.0, -1.0], 'param_1':  [1.0], 'obj_0': 0.1, 'obj_1': 0.2},
			{'param_0': [1.0, 1.0],  'param_1': [-1.0], 'obj_0': 0.2, 'obj_1': 0.1},
		]

	gryffin = Gryffin()
	samples = gryffin.recommend(observations = observations)
	print(samples)



#!/usr/bin/env python
  
__author__ = 'Florian Hase'

#=======================================================================

import copy
import json
import numpy as np 

from utilities import Logger

#=======================================================================

class DB_Writer(Logger):

	def __init__(self, config):
		self.config = config
		Logger.__init__(self, 'DB_Writer', self.config.get('verbosity'))


	def write(self, db_content, outfile, out_format):

		db_clean = {}
		for key in db_content:

			if key in ['config']: 

				db_clean[key] = {}
				db_clean[key]['general']    = db_content[key]['general']
				db_clean[key]['database']   = db_content[key]['database']

				# list-ify parameters
				db_clean[key]['parameters'] = {}
				for dict_key in db_content[key]['parameters']:
					if dict_key == 'specifics': 
						db_clean[key]['parameters'][dict_key] = []
						for element in db_content[key]['parameters'][dict_key]:
							if 'category_details' in element:
								db_clean[key]['parameters'][dict_key].append({'category_details': element['category_details']})
						continue
					db_clean[key]['parameters'][dict_key] = db_content[key]['parameters'][dict_key]

				# list-ify objectives
				db_clean[key]['objectives'] = {dict_key: db_content[key]['objectives'][dict_key].tolist() for dict_key in db_content[key]['objectives']}
				continue

			if key in ['descriptor_summary']:
				
				summary       = db_content[key][-1]
				clean_summary = {}

				name       = '-1'
				name_index = 0
				for feature_index, feature_name in enumerate(self.config.feature_names):
					if feature_name != name:
						name = feature_name
						name_index = 0

					sum_dict = summary['feature_%d' % feature_index]
					for key, value in sum_dict.items():
						sum_dict[key] = {inner_key: value[inner_key].tolist() for inner_key in value.keys()}
	
					key_name = '%s (%d)' % (name, name_index)
					clean_summary[key_name] = copy.deepcopy(sum_dict)
					name_index += 1
	
				db_clean['descriptor_summary'] = clean_summary
				continue

			values = db_content[key]
			if isinstance(values, list):
				first_element = values[0]
				if isinstance(first_element, np.ndarray):
					new_values = []
					for element in values:
						if len(element.shape) == 0:
							new_values.append(element.item())
						else:
							new_values.append([atom_element.item() for atom_element in element])				
				else:
					new_values = values
			else: 
				continue
			db_clean[key] = new_values


		formatted_string = json.dumps(db_clean, indent = 4, separators = (',', ': '))
		content = open(outfile, 'w')
		content.write(formatted_string)
		content.close()




#!/usr/bin/env python 
  
__author__ = 'Florian Hase'

#========================================================================

import numpy as np 

from gryffin.utilities import Logger
from gryffin.utilities import GryffinUnknownSettingsError

#========================================================================

class ParameterOptimizer(Logger):

	def __init__(self, config):
		self.config = config
		Logger.__init__(self, 'ParamOptimizer', verbosity = self.config.get('verbosity'))

		# parse positions
		self.pos_continuous = np.full(self.config.num_features, False, dtype = bool)
		self.pos_categories = np.full(self.config.num_features, False, dtype = bool)
		self.pos_discrete   = np.full(self.config.num_features, False, dtype = bool)
		for feature_index, feature_type in enumerate(self.config.feature_types):
			if feature_type == 'continuous':
				self.pos_continuous[feature_index] = True
			elif feature_type == 'categorical':
				self.pos_categories[feature_index] = True
			elif feature_type == 'discrete':
				self.pos_discrete[feature_index] = True
			else:
				feature_name = self.config.feature_names[feature_index]
				GryffinUnknownSettingsError('did not understand parameter type "%s" for parameter "%s".\n\t(%s) Please choose from "continuous" or "categorical"' % (feature_type, feature_name, self.template))


		# set up continuous optimization algorithms
		cont_opt_name = self.config.get('continuous_optimizer')
		if cont_opt_name == 'adam':
			from .numpy_optimizers import AdamOptimizer
			self.opt_con = AdamOptimizer()
		else:
			GryffinUnkownSettingsError('did not understand continuous optimizer "%s".\n\tPlease choose from "adam"' % cont_opt_name)

		# set up discrete optimization algorithms
		disc_opt_name = self.config.get('discrete_optimizer')
		if disc_opt_name == 'naive':
			from .numpy_optimizers import NaiveDiscreteOptimizer
			self.opt_dis = NaiveDiscreteOptimizer()
		else:
			GryffinUnknownSettingsError('did not understand discrete optimizer "%s".\n\tPlease choose from "naive"' % disc_opt_name)

		# set up categorical optimization algorithms
		cat_opt_name = self.config.get('categorical_optimizer')
		if cat_opt_name == 'naive':
			from .numpy_optimizers import NaiveCategoricalOptimizer
			self.opt_cat = NaiveCategoricalOptimizer()
		else:
			GryffinUnkownSettingsError('did not understand categorical optimizer "%s".\n\tPlease choose from "naive"' % cat_opt_name)



	def within_bounds(self, sample):
		return not (np.any(sample < self.config.feature_lowers) or np.any(sample > self.config.feature_uppers))

	def optimize_continuous(self, sample):
		proposal = self.opt_con.get_update(sample)
		if self.within_bounds(proposal):
			return proposal
		else:
			return sample

	def optimize_discrete(self, sample):
		proposal = self.opt_dis.get_update(sample)
		return proposal

	def optimize_categorical(self, sample):
		proposal = self.opt_cat.get_update(sample)
		return proposal

	
	def set_func(self, kernel, ignores = None):
		pos_continuous = self.pos_continuous.copy()
		pos_discrete   = self.pos_discrete.copy()
		pos_categories = self.pos_categories.copy()
		if ignores is not None:
			for ignore_index, ignore in enumerate(ignores):
				if ignore:
					pos_continuous[ignore_index] = False
					pos_discrete[ignore_index]   = False
					pos_categories[ignore_index] = False

		self.opt_con.set_func(kernel, pos = np.arange(self.config.num_features)[pos_continuous])
		self.opt_dis.set_func(kernel, pos = np.arange(self.config.num_features)[pos_discrete],   highest = self.config.feature_sizes[self.pos_discrete])
		self.opt_cat.set_func(kernel, pos = np.arange(self.config.num_features)[pos_categories], highest = self.config.feature_sizes[self.pos_categories])


	def optimize(self, kernel, sample, max_iter = 10):
		self.kernel = kernel

		if not self.within_bounds(sample):
			sample = np.where(sample < self.config.feature_lowers, self.config.feature_lowers, sample)
			sample = np.where(sample > self.config.feature_uppers, self.config.feature_uppers, sample)
			sample = sample.astype(np.float32)

		# update all optimization algorithms
		import time
		start = time.time()
		sample_copy = sample.copy()
		optimized   = sample.copy()
		for num_iter in range(max_iter):

			# one step of continuous
			if np.any(self.pos_continuous):
				optimized = self.optimize_continuous(optimized)

			# one step of categorical perturbation
			if np.any(self.pos_categories):
				optimized = self.optimize_categorical(optimized)

			# one step of discrete optimization
			if np.any(self.pos_discrete):
				optimized = self.optimize_discrete(optimized)

			# check for convergence
			if np.any(self.pos_continuous) and np.linalg.norm(sample_copy - optimized) < 1e-7:
				break
			else:
				sample_copy = optimized.copy()

		return optimized
	



#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

import os, sys
import pickle
import numpy      as np
import tensorflow as tf

#========================================================================

class Generator(object):

	eta      = 1e-3
	max_iter = 10**3

	def __init__(self, config_file):
		self.config_file = config_file
		with open(self.config_file, 'rb') as content:
			self.config = pickle.load(content)
		for key, value in self.config.items():
			setattr(self, key, value)

	def construct_comp_graph(self):

		tf.compat.v1.reset_default_graph()
		tf.compat.v1.disable_eager_execution()
		self.tf_descs = tf.compat.v1.placeholder(tf.float32, [None, self.num_descs])
		self.tf_objs  = tf.compat.v1.placeholder(tf.float32, [None, 1])
		
		with tf.name_scope('auto_desc_gen'):

			self.weights_0  = tf.compat.v1.get_variable('weights_0', [self.num_descs, self.num_descs], initializer = tf.initializers.identity())
			self.biases_0   = tf.compat.v1.get_variable('biases_0',  [self.num_descs],                 initializer = tf.initializers.zeros())

			self.weights_0  = self.weights_0 + tf.random.normal([self.num_descs, self.num_descs], 0., 1e-5)
			self.biases_0   = self.biases_0  + tf.random.normal([self.num_descs], 0., 1e-5)

			activation = lambda x: tf.nn.softsign(x)
			regressor  = lambda x: activation( tf.matmul(x, self.weights_0) + self.biases_0 )

			gen_descs      = regressor(self.tf_descs)
			self.gen_descs = gen_descs

			# compute correlation coefficients between descriptors and objectives
			gen_descs_mean, gen_descs_var = tf.nn.moments(gen_descs,    axes = 0)
			objs_mean,      objs_var      = tf.nn.moments(self.tf_objs, axes = 0)

			gen_descs_var += 1e-6
			objs_var      += 1e-6

			numerator   = tf.reduce_mean( (self.tf_objs - objs_mean) * (gen_descs - gen_descs_mean), axis = 0 )
			denominator = tf.sqrt( gen_descs_var * objs_var )
			corr_coeffs = numerator / denominator
			self.corr_coeffs = corr_coeffs

			# compute correlation coefficients among descriptors
			gen_descs_expand    = tf.expand_dims(gen_descs - gen_descs_mean, -1)
			gen_descs_transpose = tf.transpose(gen_descs_expand, perm = [0, 2, 1])

			gen_descs_var_expand    = tf.expand_dims(gen_descs_var, -1)
			gen_descs_var_transpose = tf.transpose(gen_descs_var_expand, perm = [1, 0])

			cov_gen_descs  = tf.reduce_mean( tf.matmul(gen_descs_expand, gen_descs_transpose), axis = 0 )
			cov_gen_descs /= tf.sqrt( tf.matmul(gen_descs_var_expand, gen_descs_var_transpose) )
			self.cov_gen_descs = cov_gen_descs

			# compute loss for deviating from target binary matrix 
#			min_corr         = 2. * 2. / (3. * np.sqrt(self.num_samples - 1) )    # corresponds to 95 % confidence interval
			min_corr         = 1. / np.sqrt(self.num_samples - 2)
			self.min_corr    = min_corr
			norm_corr_coeffs = tf.nn.leaky_relu( (tf.abs(corr_coeffs) - min_corr) / (1. - min_corr), 0.01 )
		
			loss_0  = tf.reduce_mean( tf.square( tf.sin(np.pi * norm_corr_coeffs) ) )
			loss_1  = (1. - tf.reduce_max(tf.abs(norm_corr_coeffs)))
#			loss_1  = tf.square( tf.cos( np.pi * norm_corr_coeffs[0] ) ) + tf.square( tf.cos( corr_coeffs[0] * np.pi / 2. ) )
#           loss_1 /= self.num_descs

			# compute loss for non-zero correlations in generated descriptors
			norm_cov_x = tf.nn.leaky_relu( (tf.abs(cov_gen_descs) - min_corr) / (1. - min_corr), 0.01 )
			loss_2     = tf.reduce_sum( tf.square( tf.sin(np.pi * norm_cov_x / 2.) ) ) / (self.num_descs**2 - self.num_descs)
			
			# weight regularization
			loss_3 = 1e-2 * tf.reduce_mean(tf.abs(self.weights_0))		
	
			self.loss_0 = loss_0
			self.loss_1 = loss_1
			self.loss_2 = loss_2
			self.loss_3 = loss_3

			# register training operation
			self.loss     = loss_0 + loss_1 + loss_2 + loss_3
			optimizer     = tf.compat.v1.train.AdamOptimizer(learning_rate = self.eta)
			self.train_op = optimizer.minimize(self.loss)

#		print('initializing graph variables')
		init_op   = tf.group( tf.compat.v1.global_variables_initializer(), tf.compat.v1.local_variables_initializer() )
		config    = tf.compat.v1.ConfigProto(inter_op_parallelism_threads = 1, intra_op_parallelism_threads = 1)
		self.sess = tf.compat.v1.Session(config = config)
#		print('running init_op')
		self.sess.run(init_op)	
#		print('completed init_op')



	def generate_descriptors(self):
		
#		print('OBJS',  self.objs)
#		print('DESCS', self.descs)

		self.construct_comp_graph()
		for epoch in range(self.max_iter):
			current_loss, loss_0, loss_1, loss_2, loss_3, gen_descs, _ = self.sess.run(
					[self.loss, self.loss_0, self.loss_1, self.loss_2, self.loss_3, self.weights_0, self.train_op], 
					feed_dict = {self.tf_descs: self.descs, self.tf_objs: self.objs}
				)

		corr_coeffs, gen_descs_cov = self.sess.run([self.corr_coeffs, self.cov_gen_descs], 
												    feed_dict = {self.tf_descs: self.descs, self.tf_objs: self.objs})

		# register results
		self.config['min_corrs']        = self.min_corr
		self.config['comp_corr_coeffs'] = corr_coeffs
		self.config['gen_descs_cov']    = gen_descs_cov
#		print('EVALUATING WEIGHTS')
#		with self.sess.as_default():
		self.config['weights']          = self.weights_0.eval(session = self.sess)
#		print('EVALUATED WEIGHTS')

		# run prediction
		auto_gen_descs                = self.sess.run(self.gen_descs, feed_dict = {self.tf_descs: self.grid_descs})
		self.config['auto_gen_descs'] = auto_gen_descs.astype(np.float64)
	
		# compute reduced descriptors
		sufficient_desc_indices = np.where(np.abs(corr_coeffs) > self.min_corr)[0]
		if len(sufficient_desc_indices) == 0:
			sufficient_desc_indices = np.array([0])
		reduced_gen_descs = auto_gen_descs[:, sufficient_desc_indices]
		self.config['reduced_gen_descs'] = reduced_gen_descs.astype(np.float64)


		# write pickle file
		results_path     = self.config_file.split('/')
		results_path[-1] = 'completed_%s' % results_path[-1]
		results_file     = '/'.join(results_path)

		with open(results_file, 'wb') as content:
			pickle.dump(self.config, content)


#========================================================================

if __name__ == '__main__':

	import time 
	time.sleep(0.2)
	config    = sys.argv[1]
	generator = Generator(config)
	generator.generate_descriptors()





#!/usr/bin/env python 
  
__author__ = 'Florian Hase'

#========================================================================

import numpy as np 

from gryffin.utilities import Logger
from gryffin.utilities import GryffinUnknownSettingsError

#========================================================================

class RandomSampler(Logger):

	def __init__(self, config_general, config_params):
		self.config_general = config_general
		self.config_params  = config_params
		verbosity           = self.config_general.verbosity
		if 'random_sampler' in self.config_general.verbosity:
			verbosity = self.config_general.verbosity['random_sampler']
		Logger.__init__(self, 'RandomSampler', verbosity)

		if self.config_general.sampler == 'sobol':
			from .sobol   import SobolContinuous
			from .uniform import UniformCategorical, UniformDiscrete
			self.continuous_sampler  = SobolContinuous()
			self.categorical_sampler = UniformCategorical()
			self.discrete_sampler    = UniformDiscrete()
		elif self.config_general.sampler == 'uniform':
			from .uniform import UniformCategorical, UniformContinuous, UniformDiscrete
			self.continuous_sampler  = UniformContinuous()
			self.categorical_sampler = UniformCategorical()
			self.discrete_sampler    = UniformDiscrete()
		else:
			GryffinUnknownSettingsError('did not understanding sampler setting: "%s".\n\tChoose from "uniform" or "sobol"' % self.config_general.sampler)


	def draw(self, num = 1):
		samples = []
		for param_index, param_settings in enumerate(self.config_params):
			specs = param_settings['specifics']
			if param_settings['type'] == 'continuous':
				sampled_values = self.continuous_sampler.draw(specs['low'], specs['high'], (num, param_settings['size']))
				samples.append(sampled_values)
			elif param_settings['type'] == 'categorical':
				sampled_values = self.categorical_sampler.draw(len(specs['options']), (num, param_settings['size']))
				samples.append(sampled_values)
			elif param_settings['type'] == 'discrete':
				sampled_values = self.discrete_sampler.draw(specs['low'], specs['high'], (num, param_settings['size']))
				samples.append(sampled_values)
			else:
				GryffinUnknownSettingsError('did not understand parameter type: "%s" of parameter "%s".\n\t(%s) Choose from "continuous", "discrete" or "categorical"' % (param_settings['type'], self.config_params[param_index]['name'], self.template))
		samples = np.concatenate(samples, axis = 1)
		self.log('generated uniform samples: \n%s' % str(samples), 'DEBUG')
		return samples


	def perturb(self, pos, num = 1, scale = 0.05):
		samples = []
		for param_index, param_settings in enumerate(self.config_params):
			specs = param_settings['specifics']
			if param_settings['type'] == 'continuous':
				sampled_values  = self.continuous_sampler.draw(-scale, scale, (num, param_settings['size']))
				sampled_values *= specs['high'] - specs['low']
				close_samples   = pos[param_index] + sampled_values
				close_samples   = np.where(close_samples < specs['low'],  specs['low'],  close_samples)
				close_samples   = np.where(close_samples > specs['high'], specs['high'], close_samples)
				samples.append(close_samples)
			elif param_settings['type'] in ['categorical', 'discrete']:
				sampled_values = pos[param_index] * np.ones((num, param_settings['size'])).astype(np.float32)
				samples.append(sampled_values)
			else:
				GryffinUnknownSettingsError('did not understand settings')
		samples = np.concatenate(samples, axis = 1)
		return samples


	def normal_samples(self, loc = 0., scale = 1., num = 1):
		samples = []
		for param_index, param_settings in enumerate(self.config_params):
			specs = param_settings['specifics']
			if param_settings['type'] == 'continuous':
				param_range = specs['high'] - specs['low']
				sampled_values = np.random.normal(0., scale * param_range, (num, param_settings['size'])) + loc[param_index]
				samples.append(sampled_values)
			elif param_settings['type'] == 'categorical':
				sampled_values = self.categorical_sampler.draw(len(specs['options']), (num, param_settings['size']))
				samples.append(sampled_values)
			elif param_settings['type'] == 'discrete':
				sampled_values = self.discrete_sampler.draw(specs['low'], specs['high'], (num, param_settings['size']))
				samples.append(sampled_values)
			else:
				GryffinUnknownSettingsError('did not understand variable type: "%s" of parameter "%s".\n\t(%s) Choose from "continuous", "discrete" or "categorical"' % (param_settings['type'], param_settings['name'], self.template))
		samples = np.concatenate(samples, axis = 1)
		return samples				




#!/usr/bin/env 

__author__ = 'Florian Hase'

#========================================================================

import sys
import json
import numpy as np 

from . import GryffinParseError, GryffinValueError
from . import Logger

from . import ParserJSON, CategoryParser
from . import default_general_configurations
from . import default_database_configurations
from . import safe_execute

#========================================================================

class Configuration(object):

	def __init__(self, me = ''):
		self.me          = me + ':'
		self.added_props = []
		self.added_attrs = []

	def __str__(self):
		new_line = '%s\n' % self.me
		for prop in sorted(self.added_props):
			new_line += '--> %s:\t%s\n' % (prop, getattr(self, prop))
		return(new_line)


	def __iter__(self):
		for _ in range(self.num_elements):
			info_dict = {}
			for prop_index, prop in enumerate(self.added_props):
				info_dict[prop] = self.added_attrs[prop_index][_]
			yield info_dict


	def __getitem__(self, index):
		info_dict = {}
		for prop_index, prop in enumerate(self.added_props):
			info_dict[prop] = self.added_attrs[prop_index][index]
		return info_dict

	
	def to_dict(self):
		return {prop: getattr(self, prop) for prop in sorted(self.added_props)}


	def add_attr(self, prop, attr):
		setattr(self, prop, attr)
		if not prop in self.added_props:		
			self.added_props.append(prop)
			self.added_attrs.append(attr)
		try:
			self.num_elements = len(attr)
		except TypeError:
			pass

	def get_attr(self, prop):
		return getattr(self, prop)

#========================================================================

class ConfigParser(Logger):

	TYPE_ATTRIBUTES = {
			'continuous':  ['low', 'high'],
			'discrete':    ['low', 'high', 'options', 'descriptors'],
			'categorical': ['options', 'descriptors', 'category_details'],		  
		}


	def __init__(self, config_file = None, config_dict = None):

		Logger.__init__(self, 'ConfigParser', verbosity = 0)
		self.category_parser = CategoryParser()
		self.config_file     = config_file
		self.config_dict     = config_dict

	def _parse_general(self, provided_settings):
		self.general = Configuration('general')
		for general_key, general_value in default_general_configurations.items():
			if general_key in provided_settings:
				general_value = provided_settings[general_key]
			if general_value in ['True', 'False']:
				general_value = general_value == 'True'
			self.general.add_attr(general_key, general_value)


	def _parse_database(self, provided_settings):
		self.database = Configuration('database')
		for general_key, general_value in default_database_configurations.items():
			if general_key in provided_settings:
				general_value = provided_settings[general_key]
			if general_value in ['True', 'False']:
				general_value = general_value == 'True'
			self.database.add_attr(general_key, general_value)
		if self.database.format in ['sqlite']:
			self.database.add_attr('has_db', True)
		else:
			self.database.add_attr('has_db', False)


	def _parse_parameters(self, provided_settings):
		self.parameters   = Configuration('parameters')
		self.features     = Configuration('features')
		self.kernels      = Configuration('kernels')
		param_configs     = {'name': [], 'type': [], 'size': [], 'specifics': [], 'process_constrained': []}
		feature_configs   = {'name': [], 'type': [], 'size': [], 'specifics': [], 'process_constrained': []}
		kernel_configs    = {'name': [], 'type': [], 'size': [], 'specifics': [], 'process_constrained': []}

		if len(provided_settings) == 0:
			self.log('need to define at least one parameter', 'FATAL')
		
		# parse parameter configuration
		for setting in provided_settings:

			size     = setting['size']
			num_cats = 1

			if 'process_constrained' in setting:
				setting['process_constrained'] = bool(setting['process_constrained'])
			else:
				setting['process_constrained'] = False
				

			if setting['type'] == 'continuous':
				# check order
				if setting['high'] <= setting['low']:
					GryffinValueError('upper limit (%f) needs to be larger than the lower limit (%f) for parameter "%s"' % (setting['high'], setting['low'], setting['name']))

			elif setting['type'] == 'discrete':
				# check order
				if setting['high'] <= setting['low']:
					GryffinValueError('upper limit (%f) needs to be larger than the lower limit (%f) for parameter "%s"' % (setting['high'], setting['low'], setting['name']))
				setting['options']     = np.arange(setting['low'], setting['high'], dtype = np.int32)
				setting['descriptors'] = np.arange(setting['low'], setting['high'], dtype = np.float64)
				setting['descriptors'] = np.reshape(setting['descriptors'], (len(setting['descriptors']), 1))	
				num_cats = len(setting['options'])

			elif setting['type'] == 'categorical':	
				if 'category_details' in setting:
					options, descriptors   = self.category_parser.parse(setting['category_details'])
					setting['options']     = options
					setting['descriptors'] = descriptors
				else:
					setting['category_details'] = None
					if not 'descriptors' in setting:
						setting['descriptors'] = None
						if self.general.auto_desc_gen:
							self.log('Automatic descriptor generation requested, but no descriptors provided for parameter "%s". Will not generate descriptors for this parameter. This messag will only be shown once.' % setting['name'], 'WARNING')
				num_cats = len(setting['options'])
			else:
				self.log('Did not understand parameter type "%s" for parameter "%s". Please choose from "continuous", "discrete" or "categorical".' % (setting['type'], setting['name']), 'FATAL')


			for key in param_configs.keys():
				if key == 'specifics':
					element = {spec_key: setting[spec_key] for spec_key in self.TYPE_ATTRIBUTES[setting['type']]}
				else:
					element = setting[key]

				param_configs[key].append(element)
				feature_configs[key].extend([element for _ in range(size)])
				kernel_configs[key].extend([element for _ in range(size * num_cats)])


		# write configuration
		for key in param_configs.keys():
			self.parameters.add_attr(key, param_configs[key])
			self.features.add_attr(key, feature_configs[key])
			self.kernels.add_attr(key, kernel_configs[key])


	def _parse_objectives(self, provided_settings):
		self.objectives = Configuration('objectives')
		obj_configs     = {'name': [], 'goal': [], 'hierarchy': [], 'tolerance': []}

		if len(provided_settings) == 0:
			self.log('need to define at least one objective', 'FATAL')

		elif len(provided_settings) == 1:
			setting = provided_settings[0]
			obj_configs['name'].append(setting['name'])
			obj_configs['goal'].append(setting['goal'])
			obj_configs['hierarchy'].append(0)
			obj_configs['tolerance'].append(0.0)

		else:	
			for setting in provided_settings:
				for key in obj_configs.keys():
					obj_configs[key].append(setting[key])

		# sort entries based on hierachy
		sort_indices = np.argsort(obj_configs['hierarchy'])		
		
		# check that hierarchies are complete
		for hier_index in range(len(sort_indices)):
			if np.abs(hier_index - obj_configs['hierarchy'][sort_indices[hier_index]]) > 1e-4:
				self.log('incorrect objective hierarchy\n\t... please sort objective hierarchies in ascending order, starting from 0', 'FATAL')
		
		# write configuration
		for key in obj_configs:
			self.objectives.add_attr(key, np.array(obj_configs[key])[sort_indices])

	@property
	def settings(self):
		settings_dict = {}
		settings_dict['general']    = self.general.to_dict()
		settings_dict['database']   = self.database.to_dict()
		settings_dict['parameters'] = self.parameters.to_dict()
		settings_dict['objectives'] = self.objectives.to_dict()
		return settings_dict

	@property
	def process_constrained(self):
		is_constrained = np.any(self.parameters.process_constrained)
		return is_constrained

	@property
	def param_names(self):
		return self.parameters.name

	@property
	def param_types(self):
		return self.parameters.type

	@property
	def param_sizes(self):
		return self.parameters.size

	@property
	def param_options(self):
		options = []
		for spec in self.parameters.specifics:
			if 'options' in spec:
				options.append(spec['options'])
			else:
				options.append(None)
		return options		


	@property
	def feature_descriptors(self):
		descriptors = []
		for spec in self.features.specifics:
			if 'descriptors' in spec:
				descriptors.append(spec['descriptors'])
			else:
				descriptors.append(None)
		return np.array(descriptors)

	@property
	def feature_process_constrained(self):
		return self.features.process_constrained


	@property
	def feature_lengths(self):
		lengths = []
		for spec in self.features.specifics:
			if 'options' in spec:
				lengths.append(len(spec['options']))
			else:
				lengths.append(spec['high'] - spec['low'])
		return np.array(lengths)

	@property
	def feature_lowers(self):
		lowers = []
		for spec in self.features.specifics:
			if 'low' in spec:
				lowers.append(spec['low'])
			else:
				lowers.append(0)
		return np.array(lowers)

	@property
	def feature_names(self):
		return self.features.name

	@property
	def feature_options(self):
		options = []
		for spec in self.features.specifics:
			if 'options' in spec:
				options.append(spec['options'])
			else:
				options.append(None)
		return options		

	@property
	def feature_ranges(self):
		return self.feature_uppers - self.feature_lowers

	@property
	def feature_sizes(self):
		sizes = []
		for feature_index, feature_type in enumerate(self.feature_types):
			if feature_type == 'categorical':
				feature_size = len(self.features.specifics[feature_index]['options'])
			elif feature_type == 'discrete':
				feature_size = len(self.features.specifics[feature_index]['options'])
			else:
				feature_size = 1
			sizes.append(feature_size)
		return np.array(sizes)

	@property
	def feature_types(self):
		return self.features.type

	@property
	def feature_uppers(self):
		uppers = []
		for spec in self.features.specifics:
			if 'high' in spec:
				uppers.append(spec['high'])
			else:
				uppers.append(1.)
		return np.array(uppers)

	@property
	def num_features(self):
		return len(self.feature_names)


	@property
	def kernel_lowers(self):
		lowers = []
		for spec in self.kernels.specifics:
			if 'options' in spec:
				lowers.append(0.)
			else:
				lowers.append(spec['low'])
		return np.array(lowers)

	@property
	def kernel_names(self):
		return self.kernels.name

	@property
	def kernel_ranges(self):
		return self.kernel_uppers - self.kernel_lowers

	@property
	def kernel_sizes(self):
		sizes = []
		for kernel_index, kernel_type in enumerate(self.kernel_types):
			if kernel_type == 'categorical':
				kernel_size = len(self.kernels.specifics[kernel_index]['options'])
			elif kernel_type == 'discrete':
				kernel_size = len(self.kernels.specifics[kernel_index]['options'])
			elif kernel_type == 'continuous':
				kernel_size = 1
			else:
				raise NotImplementedError()
			sizes.append(kernel_size)
		return np.array(sizes)

	@property
	def kernel_types(self):
		return self.kernels.type

	@property
	def kernel_uppers(self):
		uppers = []
		for spec in self.kernels.specifics:
			if 'options' in spec:
				uppers.append(1.)
			else:
				uppers.append(spec['high'])
		return np.array(uppers)


	@property
	def obj_names(self):
		return self.objectives.name

	@property
	def obj_tolerances(self):
		return self.objectives.tolerance
		
	@property
	def obj_goals(self):
		return self.objectives.goal

	def get(self, attr):
		return self.general.get_attr(attr)

	def get_db(self, attr):
		return self.database.get_attr(attr)

	def set_home(self, home_path):
		self.general.add_attr('home', home_path)


	def _parse(self, config_dict):
		self.config = config_dict
		
		if 'general' in self.config:
			self._parse_general(self.config['general'])
		else:
			self._parse_general({})
		self.update_verbosity(self.general.verbosity)

		if 'database' in self.config:
			self._parse_database(self.config['database'])
		else:
			self._parse_database({})

		self._parse_parameters(self.config['parameters'])
		self._parse_objectives(self.config['objectives'])


	@safe_execute(GryffinParseError)
	def parse_config_file(self, config_file = None):

		if not config_file is None:
			self.config_file = config_file

		self.json_parser = ParserJSON(json_file = self.config_file)
		self.config_dict = self.json_parser.parse()
		self._parse(self.config_dict)


	@safe_execute(GryffinParseError)
	def parse_config_dict(self, config_dict = None):
	
		if not config_dict is None:
			self.config_dict = config_dict

		self._parse(self.config_dict)


	def parse(self):
		# test if both dict and file have been provided
		if self.config_dict is not None and self.config_file is not None:
			self.log('Found both configuration file and configuration dictionary. Will parse configuration from dictionary and ignore file', 'WARNING')
			self.parse_config_dict(self.config_dict)
		elif self.config_dict is not None:
			self.parse_config_dict(self.config_dict)
		elif self.config_file is not None:
			self.parse_config_file(self.config_file)
		else:
			self.log('Cannot parse configuration due to missing configuration file or configuration dictionary', 'ERROR')



	



#!/usr/bin/env python
  
__author__ = 'Florian Hase'

#=======================================================================

import numpy as np 

from . import Chimera
from gryffin.utilities import Logger
from gryffin.utilities import GryffinUnknownSettingsError

#=======================================================================

class ObservationProcessor(Logger):

	def __init__(self, config):
		self.config  = config
		self.chimera = Chimera(self.config.obj_tolerances, self.config.get('softness'))
		Logger.__init__(self, 'ObservationProcessor', verbosity = self.config.get('verbosity'))
		
		# compute some boundaries
		self.feature_lowers = self.config.feature_lowers
		self.feature_uppers = self.config.feature_uppers
		self.soft_lower     = self.feature_lowers + 0.1 * (self.feature_uppers - self.feature_lowers)
		self.soft_upper     = self.feature_uppers - 0.1 * (self.feature_uppers - self.feature_lowers)


	def adjust_objectives(self, objs):
		'''adjust objectives based on optimization goal'''
		optim_goals   = self.config.obj_goals	
		adjusted_objs = np.empty(objs.shape)
		for obj_index, obj_goal in enumerate(optim_goals):
			if obj_goal == 'minimize':
				adjusted_objs[:, obj_index] =   objs[:, obj_index]
			elif obj_goal == 'maximize':
				adjusted_objs[:, obj_index] = - objs[:, obj_index]
			else:
				GryffinUnknownSettingsError('did not understand objective goal: "%s" for objective "%s".\n\tChoose from "minimize" or "maximize"' % (obj_goal, self.config.obj_names[obj_index]))
		return adjusted_objs


	def mirror_parameters(self, param_vector):
		# get indices
		lower_indices_prelim = np.where(param_vector < self.soft_lower)[0]	
		upper_indices_prelim = np.where(param_vector > self.soft_upper)[0]
	
		lower_indices, upper_indices = [], []
		for feature_index, feature_type in enumerate(self.config.feature_types):
			if feature_type != 'continuous': continue
			if feature_index in lower_indices_prelim:
				lower_indices.append(feature_index)
			if feature_index in upper_indices_prelim:
				upper_indices.append(feature_index)

		index_dict    = {index: 'lower' for index in lower_indices}
		for index in upper_indices:
			index_dict[index] = 'upper'

		# mirror param
		params = []
		index_dict_keys, index_dict_values = list(index_dict.keys()), list(index_dict.values())
		for index in range(2**len(index_dict)):
			param_copy = param_vector.copy()
			for jndex in range(len(index_dict)):
				if (index // 2**jndex) % 2 == 1:
					param_index = index_dict_keys[jndex]
					if index_dict_values[jndex] == 'lower':
						param_copy[param_index] = self.feature_lowers[param_index] - (param_vector[param_index] - self.feature_lowers[param_index])
					elif index_dict_values[jndex] == 'upper':
						param_copy[param_index] = self.feature_uppers[param_index] + (self.feature_uppers[param_index] - param_vector[param_index])
			params.append(param_copy)
		if len(params) == 0:
			params.append(param_vector.copy())
		return params


	def scalarize_objectives(self, objs):
		scalarized = self.chimera.scalarize(objs)
		min_obj, max_obj = np.amin(scalarized), np.amax(scalarized)
		if min_obj != max_obj:
			scaled_obj = (scalarized - min_obj) / (max_obj - min_obj)
			scaled_obj = np.sqrt(scaled_obj)
		else:
			scaled_obj = scalarized - min_obj
		return scaled_obj



	def process(self, obs_dicts):

		param_names   = self.config.param_names
		param_options = self.config.param_options
		param_types   = self.config.param_types

		# get raw results
		raw_params, raw_objs = [], []
		for obs_dict in obs_dicts:
			
			# get param-vector
			param_vector = []
			for param_index, param_name in enumerate(param_names):
				
				param_type = param_types[param_index]
				if param_type == 'continuous':
					obs_param = obs_dict[param_name]
				elif param_type == 'categorical':
					obs_param = np.array([param_options[param_index].index(element) for element in obs_dict[param_name]])
				elif param_type == 'discrete':
					obs_param = np.array([list(param_options[param_index]).index(element) for element in obs_dict[param_name]])
				param_vector.extend(obs_param)
	
			mirrored_params = self.mirror_parameters(param_vector)

			# get obj-vector
			obj_vector = np.array([obs_dict[obj_name] for obj_name in self.config.obj_names])

			# add processed params
			for param in mirrored_params:
				raw_params.append(param)
				raw_objs.append(obj_vector)

		raw_objs, raw_params = np.array(raw_objs), np.array(raw_params)

		params = raw_params	
		adjusted_objs = self.adjust_objectives(raw_objs)
		objs          = self.scalarize_objectives(adjusted_objs)

		return params, objs





#!/usr/bin/env python

__author__ = 'Florian Hase'

#========================================================================

import sys
import numpy as np 

from gryffin.utilities import GryffinModuleError

try:
	import sobol_seq
except ModuleNotFoundError:
	_, error_message, _ = sys.exc_info()
	extension = '\n\tTry installing the sobol_seq package or use uniform sampling instead.'
	GryffinModuleError(str(error_message) + extension)

#========================================================================

class SobolCategorical(object):

	def __init__(self):
		pass

#========================================================================


class SobolContinuous(object):

	def __init__(self, seed = None):
		if seed is None:
			seed = np.random.randint(low = 0, high = 10**5)
		self.seed = seed


	def draw(self, low, high, size):
		num, dim = size[0], size[1]
		samples = []
		for _ in range(num):
			vector, seed = sobol_seq.i4_sobol(dim, self.seed)
			sample = (high - low) * vector + low
			self.seed = seed
			samples.append(sample)
		return np.array(samples)


#========================================================================





#!/usr/bin/env python 
  
__author__ = 'Florian Hase'

#========================================================================

import numpy as np 

from . import AbstractOptimizer

#========================================================================

class AdamOptimizer(AbstractOptimizer):

	iterations = 0

	def __init__(self, func = None, eta = 0.1, beta_1 = 0.9, beta_2 = 0.999, epsilon = 1e-8, decay = 0., *args, **kwargs):
		AbstractOptimizer.__init__(self, func, *args, **kwargs)
		self.eta     = eta
		self.beta_1  = beta_1
		self.beta_2  = beta_2
		self.epsilon = epsilon
		self.decay   = decay
		self.initial_decay = self.decay
		self.ms, self.vs = None, None


	def set_func(self, func, pos = None):
		self.iterations = 0
		self._set_func(func, pos)
		if hasattr(self, 'ms'):
			delattr(self, 'ms')
			delattr(self, 'vs')

	def get_update(self, vector):
		grads = self.grad(vector)
		eta   = self.eta
		if self.initial_decay > 0.:
			eta *= (1. / (1. + self.decay * self.iterations))
		
		next_iter = self.iterations + 1
		eta_next  = eta * (np.sqrt(1. - np.power(self.beta_2, next_iter)) / (1. - np.power(self.beta_1, next_iter)))
		
		if not hasattr(self, 'ms'):
			self.ms = [0. for element in vector]
			self.vs = [0. for element in vector]

		update = np.zeros(len(vector)).astype(np.float32) + np.nan
		for index, element, grad, mass, vel in zip(range(len(vector)), vector, grads, self.ms, self.vs):
#			print('...', index, update, grad)
			m_next = (self.beta_1 * mass) + (1. - self.beta_1) * grad
			v_next = (self.beta_2 * vel)  + (1. - self.beta_2) * np.square(grad)
			p_next = element - eta_next * m_next / (np.sqrt(v_next) + self.epsilon)
			self.ms[index] = m_next
			self.vs[index] = v_next
			update[index]  = p_next

		self.iterations += 1
		return np.array(update)

#========================================================================

if __name__ == '__main__':

	import matplotlib.pyplot as plt 
	import seaborn as sns 

	adam = AdamOptimizer()

	def func(x):
		return (x - 1)**2

	adam.set_func(func, pos = np.arange(1))

	domain = np.linspace(-1, 3, 200)
	values = func(domain)

	start = np.zeros(1) - 0.8

	plt.ion()

	for _ in range(10**3):
		
		plt.clf()
		plt.plot(domain, values)
		plt.plot(start, func(start), marker = 'o', color = 'k')

		start = adam.get_update(start)

		plt.pause(0.05)

#		print(start, func(start))







#!/usr/bin/env python

__author__ = 'Florian Hase'

#========================================================================

import numpy as np

#========================================================================

class UniformCategorical(object):
	
	def __init__(self):
		pass

	def draw(self, num_options, size):
		return np.random.choice(num_options, size = size).astype(np.float32)

#========================================================================


class UniformContinuous(object):
	
	def __init__(self):
		pass

	def draw(self, low, high, size):
		return np.random.uniform(low = low, high = high, size = size).astype(np.float32)

#========================================================================


class UniformDiscrete(object):

	def __init__(self):
		pass

	def draw(self, low, high, size):
		return np.random.randint(low = 0, high = high - low, size = size).astype(np.float32)





#!/usr/bin/env python 
  
__author__ = 'Florian Hase'

#========================================================================

import numpy as np

from . import AbstractOptimizer

#========================================================================

class NaiveCategoricalOptimizer(AbstractOptimizer):


	def __init__(self, func = None, *args, **kwargs):
		AbstractOptimizer.__init__(self, func, *args, **kwargs)
		

	def set_func(self, func, pos = None, highest = None):
		# set function
		self.highest = highest
		self._set_func(func, pos)


	def get_update(self, vector):
		func_best     = self.func(vector) 
		for pos_index, pos in enumerate(self.pos):
			if pos is None: continue
		
			current = vector[pos]
			perturb = np.random.choice(self.highest[pos_index])
			vector[pos] = perturb

			# cache the response!
#			if self.only_cats: 
#				func_cand = self.get_response(vector)
#			else:
			func_cand = self.func(vector)

			if func_cand < func_best:
				func_best = func_cand
			else:
				vector[pos] = current			
		return vector


	def old_get_update(self, vector):
		func_best     = self.func(vector) 
		for pos_index in range(self.num_pos):
			if self.pos[pos_index] is None: continue
		
			current = vector[pos_index]
			perturb = np.random.choice(self.highest[pos_index])
			vector[pos_index] = perturb

			# cache the response!
#			if self.only_cats: 
#				func_cand = self.get_response(vector)
#			else:
			func_cand = self.func(vector)

			if func_cand < func_best:
				func_best = func_cand
			else:
				vector[pos_index] = current			
		return vector

#========================================================================

if __name__ == '__main__':

	import matplotlib.pyplot as plt 
	import seaborn as sns

	colors = sns.color_palette('YlOrRd', 256)

	cat_opt = NaiveCategoricalOptimizer()
	highest = 20
	max_loss = np.sum(np.square(np.zeros(2) - highest / 2.))
	def func(array):
		loss = np.sum( np.square( array - highest / 2. ) ) / max_loss
		return loss

	cat_opt.set_func(func, highest = [highest, highest], pos = np.arange(2))

	lines    = np.arange(highest + 2) - 0.5	

	start = np.zeros(2)

	fig = plt.figure(figsize = (6, 6))

	for _ in range(10**3):

		plt.clf()

		plt.plot(start[0], start[1], marker = 'o', color = 'k', markersize = 10)

		for x_line in lines:
			plt.plot([lines[0], lines[-1]], [x_line, x_line], color = 'k', lw = 0.5)
			plt.plot([x_line, x_line], [lines[0], lines[-1]], color = 'k', lw = 0.5)

		for x_element in np.arange(highest + 1):
			for y_element in np.arange(highest + 1):
				array = np.array([x_element, y_element])
				loss  = func(array)
				color_index = int( 255 * loss  )
				plt.plot(x_element, y_element, color = colors[color_index], marker = 's', markersize = 10, alpha = 0.5)

		start = cat_opt.get_update(start)

		plt.pause(0.05)





#!/usr/bin/env 

__author__ = 'Florian Hase'

#========================================================================

import sys 

from gryffin.utilities import GryffinModuleError, GryffinVersionError

#========================================================================

try:
	import tensorflow as tf
except ModuleNotFoundError:
	_, error_message, _ = sys.exc_info()
	extension = '\n\tAutomatic descriptor generation requires the tensorflow package. Please install tensorflow or disable automatic descriptor generation by setting auto_desc_gen="False".'
	GryffinModuleError(str(error_message) + extension)

#========================================================================

from .descriptor_generator import DescriptorGenerator



#!/usr/bin/env python 
  
__author__ = 'Florian Hase'

#========================================================================

import numpy as np 
import multiprocessing
from multiprocessing import Manager, Process

from gryffin.utilities import Logger

#========================================================================

class SampleSelector(Logger):

	def __init__(self, config):
		self.config = config
		Logger.__init__(self, 'SampleSelector', verbosity = self.config.get('verbosity'))
		self.num_cpus = multiprocessing.cpu_count()


	def compute_exp_objs(self, proposals, kernel_contribution, batch_index, return_index, result_dict = None):

		samples  = proposals[batch_index]
		exp_objs = np.empty(len(samples))

		for sample_index, sample in enumerate(samples):
			num, inv_den = kernel_contribution(sample)
			kernel_contrib = (num + self.sampling_param_values[batch_index]) * inv_den
			exp_objs[sample_index] = np.exp( - kernel_contrib)

		if result_dict.__class__.__name__ == 'DictProxy':
			result_dict[return_index] = exp_objs
		else:
			return exp_objs


	
	def select(self, num_samples, proposals, kernel_contribution, sampling_param_values, obs_params):

		num_obs = len(obs_params)	
		feature_ranges = self.config.feature_ranges
		char_dists     = feature_ranges / float(num_obs)**0.5
		self.sampling_param_values = sampling_param_values		
	
		# compute acq func values
		if self.config.get('parallel'):
			result_dict = Manager().dict()
			
			# get the number of splits
			num_splits = self.num_cpus // len(sampling_param_values) + 1
			split_size = proposals.shape[1] // num_splits

			processes = []
			for batch_index in range(len(sampling_param_values)):
				for split_index in range(num_splits):
					
					split_start = split_size * split_index
					split_end   = split_size * (split_index + 1)
					return_index = num_splits * batch_index + split_index
					process = Process(target = self.compute_exp_objs, args = (proposals[:, split_start : split_end], kernel_contribution, batch_index, return_index, result_dict))
					processes.append(process)
					process.start()
				for process_index, process in enumerate(processes):
					process.join()

		else:
			num_splits  = 1
			result_dict = {}
			for batch_index in range(len(sampling_param_values)):
				return_index = batch_index
				result_dict[return_index] = self.compute_exp_objs(proposals, kernel_contribution, batch_index, return_index)

		# collect results
		exp_objs = []
		for batch_index in range(len(sampling_param_values)):
			batch_exp_objs = []
			for split_index in range(num_splits):
				return_index = num_splits * batch_index + split_index

				batch_exp_objs.append(result_dict[return_index])
			exp_objs.append(np.concatenate(batch_exp_objs))
		exp_objs = np.array(exp_objs)

		# compute prior recommendation punishments
		for batch_index in range(len(sampling_param_values)):
			batch_proposals = proposals[batch_index, : exp_objs.shape[1]]

			# compute distance to each obs_param
			distances     = [ np.sum((obs_params - batch_proposal)**2, axis = 1) for batch_proposal in batch_proposals]
			distances     = np.array(distances)			
			min_distances = np.amin(distances, axis = 1)
			ident_indices = np.where(min_distances < 1e-8)[0]
		
			exp_objs[batch_index, ident_indices] = 0.


		# collect samples
		samples = []
		for sample_index in range(num_samples):
			new_samples = []

			for batch_index in range(len(sampling_param_values)):
				batch_proposals = proposals[batch_index]

				# compute diversity punishments
				div_crits = np.ones(exp_objs.shape[1])

				for proposal_index, proposal in enumerate(batch_proposals[:exp_objs.shape[1]]):
					obs_min_distance = np.amin([np.abs(proposal - x) for x in obs_params], axis = 0)
					if len(new_samples) > 0:
						min_distance = np.amin([np.abs(proposal - x) for x in new_samples], axis = 0)
						min_distance = np.minimum(min_distance, obs_min_distance)
					else:
						min_distance = obs_min_distance

					div_crits[proposal_index] = np.minimum(1., np.mean( np.exp( 2. * (min_distance - char_dists) / feature_ranges) ))					


				# reweight rewards
				reweighted_rewards   = exp_objs[batch_index] * div_crits
				largest_reward_index = np.argmax( reweighted_rewards )
				
				new_sample = batch_proposals[largest_reward_index]
				new_samples.append(new_sample)

				# update reward of selected sample
				exp_objs[batch_index, largest_reward_index] = 0.

			samples.append(new_samples)
		samples = np.concatenate(samples)
		return samples









#!/usr/bin/env python
  
__author__ = 'Florian Hase'

#=======================================================================

import pickle 

from utilities import Logger

#=======================================================================

class DB_Writer(Logger):

	def __init__(self, config):
		self.config = config
		Logger.__init__(self, 'DB_Writer', self.config.get('verbosity'))

	def write(self, db_content, outfile, out_format):
		pickle.dump(db_content, open(outfile, 'wb'))






#!/usr/bin/env python 
  
__author__ = 'Florian Hase'

#========================================================================

import numpy as np

#========================================================================

class Chimera(object):

	def __init__(self, tolerances, softness = 0.0, absolutes = None):
		self.tolerances = tolerances
		self.absolutes  = absolutes
		if absolutes is None:
			self.absolutes = np.zeros(len(tolerances)) + np.nan
		self.softness   = softness


	def soft_step(self, value):
		arg = - value / self.softness
		return 1. / (1. + np.exp(arg))


	def hard_step(self, value):
		result = np.empty(len(value))
		result = np.where(value > 0., 1., 0.)
		return result


	def step(self, value):
		if self.softness < 1e-5:
			return self.hard_step(value)
		else:
			return self.soft_step(value)


	def rescale(self, raw_objs):
		res_objs = np.empty(raw_objs.shape)
		res_abs  = np.empty(self.absolutes.shape)
		for index in range(raw_objs.shape[1]):
			min_objs, max_objs = np.amin(raw_objs[:, index]), np.amax(raw_objs[:, index])
			if min_objs < max_objs:
				res_abs[index]     = (self.absolutes[index] - min_objs) / (max_objs - min_objs)
				res_objs[:, index] = (raw_objs[:, index] - min_objs) / (max_objs - min_objs)
			else:
				res_abs[index]     = self.absolutes[index] - min_objs
				res_objs[:, index] = raw_objs[:, index] - min_objs
		return res_objs, res_abs				


	def shift_objectives(self, objs, res_abs):
		transposed_objs  = objs.transpose()
		shapes           = transposed_objs.shape
		shifted_objs     = np.empty((shapes[0] + 1, shapes[1]))
		
		mins, maxs, tols = [], [], []
		domain           = np.arange(shapes[1])
		shift            = 0
		for obj_index, obj in enumerate(transposed_objs):

			# get absolute tolerances
			minimum = np.amin(obj[domain])
			maximum = np.amax(obj[domain])
			mins.append(minimum)
			maxs.append(maximum)
			tolerance = minimum + self.tolerances[obj_index] * (maximum - minimum)
			if np.isnan(tolerance):
				tolerance = res_abs[obj_index]			

			# adjust region of interest
			interest = np.where(obj[domain] < tolerance)[0]
			if len(interest) > 0:
				domain = domain[interest]
	
			# apply shift	
			tols.append(tolerance + shift)
			shifted_objs[obj_index] = transposed_objs[obj_index] + shift

			# compute new shift
			if obj_index < len(transposed_objs) - 1:
				shift -= np.amax(transposed_objs[obj_index + 1][domain]) - tolerance
			else:
				shift -= np.amax(transposed_objs[0][domain]) - tolerance
				shifted_objs[obj_index + 1] = transposed_objs[0] + shift
		return shifted_objs, tols



	def scalarize_objs(self, shifted_objs, abs_tols):
		scalar_obj = shifted_objs[-1].copy()
		for index in range(0, len(shifted_objs) - 1)[::-1]:
			scalar_obj *= self.step( - shifted_objs[index] + abs_tols[index])
			scalar_obj += self.step(   shifted_objs[index] - abs_tols[index]) * shifted_objs[index]
		return scalar_obj.transpose()

	

	
	def scalarize(self, raw_objs):
		res_objs, res_abs      = self.rescale(raw_objs)
		shifted_objs, abs_tols = self.shift_objectives(res_objs, res_abs) 
		scalarized_obj         = self.scalarize_objs(shifted_objs, abs_tols)
		return scalarized_obj


#========================================================================




#!/usr/bin/env 

__author__ = 'Florian Hase'

#========================================================================

import pickle

#========================================================================

class ParserPickle(object):

	def __init__(self, pickle_file = None):
		self.pickle_file = pickle_file

	def parse(self, pickle_file = None):
		# update pickle file
		if not pickle_file is None:
			self.pickle_file = pickle_file

		# parse configuration
		if not self.pickle_file is None:
			with open(self.pickle_file, 'rb') as content:
				self.parsed_pickle = pickle.load(content)
		return self.parsed_pickle




#!/usr/bin/env 

__author__ = 'Florian Hase'

#========================================================================

import numpy as np

from . import ParserJSON, ParserPickle

#========================================================================

class CategoryParser(object):

	def __init__(self, file_name = None):
		self.file_name     = file_name
		self.json_parser   = ParserJSON()
		self.pickle_parser = ParserPickle()
		self.parsers       = [self.json_parser, self.pickle_parser]

	def parse(self, file_name = None):
		# update file name
		if not file_name is None:
			self.file_name = file_name

		suffix = self.file_name.split('.')[-1]
		if suffix == 'json':
			cat_details = self.json_parser.parse(self.file_name)
		elif suffix == 'pkl':
			cat_details = self.pickle_parser.parse(self.file_name)
		elif suffix == 'csv':
			raise NotImplementedError
		else:
			# try different parsers
			cat_details = None
			for parser in self.parsers:
				try:
					cat_details = parser.parse(self.file_name)
				except:
					pass
				if cat_details is not None:
					break
			else:
				PhoenicsUnkown

		# reformat information
		options     = [cat_dict['name'] for cat_dict in cat_details]
	
		try:
			descriptors = np.array([cat_dict['descriptors'] for cat_dict in cat_details])
	
			# TODO: rescale the descriptors!!
			min_descriptors, max_descriptors = np.amin(descriptors, axis = 0), np.amax(descriptors, axis = 0)
			descriptors = (descriptors - min_descriptors) / (max_descriptors - min_descriptors)

			descs = []
			for descriptor in descriptors:
				desc = descriptor[np.where(np.isfinite(descriptor))[0]]
				descs.append(desc)
			descs = np.array(descs)
			descriptors = descs

		except KeyError:
			descriptors = None
			
		return options, descriptors




#!/usr/bin/env 

__author__ = 'Florian Hase'

#========================================================================

import numpy as np 

#========================================================================

def sigmoid(x):
	return 1. / (1. + np.exp( - x))

#========================================================================

class NumpyGraph(object):

	def __init__(self, config, model_details):
		self.config = config
#		Logger.__init__(self, 'NumpyGraph', self.config.get('verbosity'))

		self.model_details = model_details
		for key, value in self.model_details.items():
			setattr(self, '_%s' % str(key), value)

		self.feature_size    = len(self.config.kernel_names)
		self.bnn_output_size = len(self.config.kernel_names)
		self.target_size     = len(self.config.kernel_names)

	
	def declare_training_data(self, features):
		
		self.num_obs  = len(features)
		self.features = features


	def compute_kernels(self, posteriors):
	
		tau_rescaling = np.zeros((self.num_obs, self.bnn_output_size))
		kernel_ranges = self.config.kernel_ranges
		for obs_index in range(self.num_obs):
			tau_rescaling[obs_index] += kernel_ranges
		tau_rescaling = tau_rescaling**2
	
		# sample from BNN
		activations = [np.tanh, np.tanh, lambda x: x]
		post_layer_outputs = [np.array([self.features for _ in range(self._num_draws)])]
		for layer_index in range(self._num_layers):

			weight     = posteriors['weight_%d' % layer_index]
			bias       = posteriors['bias_%d'   % layer_index]
			activation = activations[layer_index]

			outputs = []
			for sample_index in range(len(weight)):

				single_weight = weight[sample_index]
				single_bias   = bias[sample_index]

				output = activation( np.matmul( post_layer_outputs[-1][sample_index], single_weight) + single_bias)
				outputs.append(output)

#			print(np.array(outputs).shape)
#
#			quit()

			post_layer_output = np.array(outputs)
			post_layer_outputs.append(post_layer_output)

#			print('DONE')

		post_bnn_output = post_layer_outputs[-1]

		# note: np.random.gamma is parametrized with k and theta, while ed.models.Gamma is parametrized with alpha and beta
		post_tau_normed = np.random.gamma( self.num_obs**2 + np.zeros(post_bnn_output.shape), np.ones(post_bnn_output.shape))
		post_tau        = post_tau_normed / tau_rescaling
		post_sqrt_tau   = np.sqrt(post_tau)
		post_scale      = 1. / post_sqrt_tau

#		print('BNN OUTPUT', post_bnn_output.shape)
#		print('BNN TAU', post_sqrt_tau.shape)
#		quit()
		

		# map BNN output to predictions

		post_kernels = {}
		
		target_element_index = 0
		kernel_element_index = 0

		while kernel_element_index < len(self.config.kernel_names):

			kernel_type = self.config.kernel_types[kernel_element_index]
			kernel_size = self.config.kernel_sizes[kernel_element_index]

			feature_begin, feature_end = target_element_index, target_element_index + 1
			kernel_begin,  kernel_end  = kernel_element_index, kernel_element_index + kernel_size

#			print('KERNEL', kernel_begin, kernel_end)

			post_relevant = post_bnn_output[:, :, kernel_begin : kernel_end]

			if kernel_type == 'continuous':

				lowers       = self.config.kernel_lowers[kernel_begin : kernel_end]
				uppers       = self.config.kernel_uppers[kernel_begin : kernel_end]
				post_support = (uppers - lowers) * (1.2 * sigmoid(post_relevant) - 0.1) + lowers
				
				post_kernels['param_%d' % target_element_index] = {'loc':       post_support,
																   'sqrt_prec': post_sqrt_tau[:, :, kernel_begin : kernel_end],
																   'scale':     post_scale[:, :, kernel_begin : kernel_end]}


			elif kernel_type == 'categorical':

				post_temperature = 0.5 + 10.0 / self.num_obs
#				post_temperature = 0.4
				post_support     = post_relevant

				post_probs        = 1. / (1. + np.exp( - post_support))
				post_probs_normed = post_probs / np.sum(post_probs)

				# sample from relaxed one hot categorical
				gumbel_samples       = np.random.gumbel(loc = 0., scale = 1., size = post_support.shape)
				exp_samples          = np.exp( ( post_support + gumbel_samples) / post_temperature )
				exp_samples_sums     = np.sum(exp_samples, axis = 2)

				post_predict_relaxed = np.zeros(exp_samples.shape)
				for sample_index in range(exp_samples.shape[0]):
					for param_index in range(exp_samples.shape[1]):
						post_predict_relaxed[sample_index, param_index] = exp_samples[sample_index, param_index] / exp_samples_sums[sample_index, param_index]

				post_kernels['param_%d' % target_element_index] = {'probs': post_predict_relaxed}

			elif kernel_type == 'discrete':
				
				post_temperature = 0.5 + 1.0 / self.num_obs
	#			post_temperature = 0.4
				post_support     = post_relevant

				post_probs        = 1. / (1. + np.exp( - post_support))
				post_probs_normed = post_probs / np.sum(post_probs)

				# sample from relaxed one hot categorical
				gumbel_samples = np.random.gumbel(loc = 0., scale = 1., size = post_support.shape)
				
				exp_samples          = np.exp( ( post_support + gumbel_samples) / post_temperature )
				exp_samples_sums     = np.sum(exp_samples, axis = 2)

				post_predict_relaxed = np.zeros(exp_samples.shape)
				for sample_index in range(exp_samples.shape[0]):
					for param_index in range(exp_samples.shape[1]):
						post_predict_relaxed[sample_index, param_index] = exp_samples[sample_index, param_index] / exp_samples_sums[sample_index, param_index]

				post_kernels['param_%d' % target_element_index] = {'probs': post_predict_relaxed}



			else:
				raise NotImplementedError

			target_element_index += 1
			kernel_element_index += kernel_size


		return post_kernels






#!/usr/bin/env 
  
__author__ = 'Florian Hase'

#========================================================================

import sys

from gryffin.utilities import GryffinModuleError, GryffinVersionError

#========================================================================

try:
    import tensorflow as tf
except ModuleNotFoundError:
    _, error_message, _ = sys.exc_info()
    extension = '\n\tTry installing the tensorflow package or use a different backend instead.'
    GryffinModuleError(str(error_message) + extension)
#========================================================================


import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

from .numpy_graph      import NumpyGraph
from .tfprob_interface import TfprobNetwork




#!/usr/bin/env 

__author__ = 'Florian Hase'

#========================================================================

import time
import numpy as np 
from multiprocessing import Process, Manager

from gryffin.utilities import Logger

from .kernel_prob_reshaping import KernelReshaper

#========================================================================


class CategoryReshaper(Logger):

	def __init__(self, config):
		self.config = config
		Logger.__init__(self, 'CategoryReshaper', verbosity = self.config.get('verbosity'))

		self.kernel_reshaper = KernelReshaper()


	def cython_recompute_probs(self, cat_probs, descriptors, index, return_dict = None):
		recomputed_probs = self.kernel_reshaper.reshape_probs(cat_probs, descriptors)
		if return_dict.__class__.__name__ == 'DictProxy':
			return_dict[index] = recomputed_probs
		else:	
			return recomputed_probs


	def python_recompute_probs(self, cat_probs, descriptors, index, return_dict = None):
		'''
			shape of cat_probs    ... (# samples, # obs, # kernels)
			shape of descriptors: ... (# kernels, # descriptors)
		'''
		num_samples     = cat_probs.shape[0]
		num_obs         = cat_probs.shape[1]
		num_kernels     = cat_probs.shape[2]
		num_descriptors = descriptors.shape[1] 

		start = time.time()

		recomputed_probs = np.empty(cat_probs.shape)

		# put on separate thread
		for sample_index in range(num_samples):
			# put on separate thread
			for obs_index in range(num_obs):
				
				start_0 = time.time()

				probs = cat_probs[sample_index, obs_index]

				# compute distances to all categories
				distances = np.empty(num_kernels)
				for target_cat_index in range(num_kernels):

					ds2 = 0.

					for desc_index in range(num_descriptors):

						averaged_descriptor = 0.
						for kernel_index in range(num_kernels):
							averaged_descriptor += probs[kernel_index] * descriptors[kernel_index, desc_index]

						dyi = 0.
						for kernel_index in range(num_kernels):
							partial_contrib = descriptors[target_cat_index, desc_index] - averaged_descriptor
							dyi += partial_contrib

						ds2 += dyi * dyi

					distances[target_cat_index] = np.sqrt(ds2 / num_descriptors)
					
				# got the distances
				rescaled_probs = np.exp( - distances) / np.sum( np.exp( - distances))
				recomputed_probs[sample_index, obs_index] = rescaled_probs

				end_0 = time.time()
			
		end = time.time()
		total_time = end - start

		if return_dict.__class__.__name__ == 'DictProxy':
			return_dict[index] = recomputed_probs
		else:	
			return recomputed_probs

	
	def reshape(self, raw_probs, descriptors):
		# ... raw_probs = (# samples, # obs, # kernels)

		# assign kernels to individual parameters
		feature_probs  = []
		kernel_sizes = self.config.kernel_sizes
		kernel_start = 0
		while kernel_start < raw_probs.shape[2]:
			kernel_end = kernel_start + kernel_sizes[kernel_start]
			feature_probs.append(raw_probs[:, :, kernel_start : kernel_end])
			kernel_start = kernel_end

		# extract all relevant information for all parameters
		parsed_probs = []
		for feature_index, feature_options in enumerate(self.config.feature_options):
			feature_descriptor = descriptors[feature_index]
			parsed_prob = {'options': feature_options, 'descriptors': feature_descriptor, 
						   'probs':   feature_probs[feature_index], 'feature_index': feature_index}
			parsed_probs.append(parsed_prob)

		# run reshaping algorithm
		if self.config.get('parallel'):
			result_dict = Manager().dict()
			processes   = []
			for feature_index in range(len(parsed_probs)):
				prob_dict = parsed_probs[feature_index]
				if prob_dict['descriptors'] is None: 
					result_dict[feature_index] = prob_dict['probs']
					continue
				process   = Process(target = self.cython_recompute_probs, args = (prob_dict['probs'], prob_dict['descriptors'], feature_index, result_dict))
#				process   = Process(target = self.python_recompute_probs, args = (prob_dict['probs'], prob_dict['desriptors'], feature_index, result_dict))
				processes.append(process)
				process.start()
			for process in processes:	
				process.join()
		else:
			result_dict = {}
			for feature_index in range(len(parsed_probs)):
				prob_dict = parsed_probs[feature_index]
				if prob_dict['descriptors'] is None: 
					result_dict[feature_index] = prob_dict['probs']
					continue
				result_dict[feature_index] = self.cython_recompute_probs(prob_dict['probs'], prob_dict['descriptors'], feature_index)
#				result_dict[feature_index] = self.python_recompute_probs(prob_dict['probs'], prob_dict['descriptors'], feature_index)

		recomputed_probs = []
		for feature_index in range(len(parsed_probs)):
			if feature_index in result_dict:
				recomputed_probs.append(result_dict[feature_index])
			else:
				recomputed_probs.append(np.ones(( raw_probs.shape[0], raw_probs.shape[1], 1 )))
				
		# assemble reshaped probs
		reshaped_probs = np.concatenate(recomputed_probs, axis = 2)
		return reshaped_probs




#!/usr/bin/env 

__author__ = 'Florian Hase'

#========================================================================

import sys
import threading

#========================================================================

def safe_execute(error):
	def decorator_wrapper(function):
		def wrapper(*args, **kwargs):
			try:
				function(*args, **kwargs)
			except:
				error_type, error_message, traceback = sys.exc_info()
				error(error_message)
		return wrapper
	return decorator_wrapper



def thread(function):
	def wrapper(*args, **kwargs):
		background_thread = threading.Thread(target = function, args = args, kwargs = kwargs)
		background_thread.start()
	return wrapper





#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

import uuid
import time
import copy
import sqlalchemy as sql

from . import AddEntry, FetchEntries, UpdateEntries
from gryffin.utilities                       import Logger
from gryffin.utilities.decorators            import thread

#========================================================================

class SqliteDatabase(Logger):

	SQLITE_COLUMNS = {'float':   sql.Float(),
					  'bool':    sql.Boolean(),
					  'integer': sql.Integer(),
					  'pickle':  sql.PickleType(),
					  'string':  sql.String(512),}

	def __init__(self, path, attributes, name = 'table', verbosity = 0):

		self.WRITING_REQUESTS = []
		self.READING_REQUESTS = {}
		self.UPDATE_REQUESTS  = []
		Logger.__init__(self, 'SQLite interface', verbosity = verbosity)

		self.db_path              = 'sqlite:///%s/search_progress.db' % path
		self.attributes           = attributes
		self.name                 = name
		
		self.log('creating database %s at %s' % (self.name, self.db_path), 'DEBUG')

		# create database 
		self.db       = sql.create_engine(self.db_path)
		self.db.echo  = False		
		self.metadata = sql.MetaData(self.db)

		# create table in database
		self.table = sql.Table(self.name, self.metadata)
		for name, att_type in self.attributes.items():
			self.table.append_column(sql.Column(name, self.SQLITE_COLUMNS[att_type]))
		self.table.create(checkfirst = True)

		# start request processor
		self._process_requests()

	#====================================================================

	def _return_dict(function):
		def wrapper(self, *args, **kwargs):
			entries    = function(self, *args, **kwargs)
			info_dicts = [{key: entry[key] for key in self.attributes} for entry in entries]
			return info_dicts
		return wrapper

	#====================================================================

	@thread
	def _process_requests(self):
		self._processing_requests = True
		keep_processing           = True
		iteration_index           = 0
		while keep_processing:
			num_reading_requests = len(self.READING_REQUESTS)
			num_writing_requests = len(self.WRITING_REQUESTS)
			num_update_requests  = len(self.UPDATE_REQUESTS)

			iteration_index += 1

			# run at most one reading request
			request_keys = copy.deepcopy(list(self.READING_REQUESTS.keys()))
			for request_key in request_keys:
				if not self.READING_REQUESTS[request_key].executed:
					self.READING_REQUESTS[request_key].execute()
					break

			# run all update requests
			for update_index in range(num_update_requests):
				update_request = self.UPDATE_REQUESTS.pop()
				update_request.execute()

			# run at most one writing request
			if num_writing_requests > 0:
				writing_request = self.WRITING_REQUESTS.pop()
				writing_request.execute()

			# clean reading requests
			request_keys = copy.deepcopy(list(self.READING_REQUESTS.keys()))
			delete_keys  = []
			for request_key in request_keys:
				if self.READING_REQUESTS[request_key].entries_fetched:
					delete_keys.append(request_key)
			for request_key in delete_keys:
				del self.READING_REQUESTS[request_key]

			keep_processing = len(self.WRITING_REQUESTS) > 0 or len(self.UPDATE_REQUESTS) > 0 or len(self.READING_REQUESTS) > 0
		self._processing_requests = False

	#====================================================================


	def add(self, info_dict):
		add_entry = AddEntry(self.db, self.table, info_dict)
		self.WRITING_REQUESTS.append(add_entry)
		if not self._processing_requests:
			self._process_requests()
		


	@_return_dict
	def fetch_all(self, condition_dict):
		condition_keys   = list(condition_dict.keys())
		condition_values = list(condition_dict.values())

		# define the selection
		selection = sql.select([self.table])
		for index, key in enumerate(condition_keys):
			if isinstance(condition_values[index], list):
				# with a list, we need to combine all possibilities with _or
				if len(condition_values[index]) == 0:
					return []
				filters   = [getattr(self.table.c, key) == value for value in condition_values[index]]
				condition = sql.or_(*filters)
			else:
				condition = getattr(self.table.c, key) == condition_values[index]
			selection = selection.where(condition)

		fetch_entries = FetchEntries(self.db, self.table, selection)
		fetch_keys    = str(uuid.uuid4())
		self.READING_REQUESTS[fetch_keys] = fetch_entries
		if not self._processing_requests:
			self._process_requests()

		entries = fetch_entries.get_entries()
		self.log('fetched all information from database %s' % self.name, 'DEBUG')
		return entries




	def update_all(self, condition_dict, update_dict):
		
		condition_keys   = list(condition_dict.keys())
		condition_values = list(condition_dict.values())
		
		# defining the selection 
		update = sql.update(self.table).values(update_dict).where(getattr(self.table.c, condition_keys[0]) == condition_values[0])
		for index, key in enumerate(condition_keys[1:]):
			update = update.where(getattr(self.table.c, key) == condition_values[index + 1])

		# submitting the update
		update_entries = UpdateEntries(self.db, self.table, update)
		self.WRITING_REQUESTS.append(update_entries)
		if not self._processing_requests:
			self._process_requests()








from .decorators      import safe_execute
from .defaults        import default_general_configurations
from .defaults        import default_database_configurations

from .exceptions      import GryffinParseError
from .exceptions      import GryffinModuleError
from .exceptions      import GryffinNotFoundError
from .exceptions      import GryffinUnknownSettingsError
from .exceptions      import GryffinValueError
from .exceptions      import GryffinVersionError

from .logger          import Logger

from .json_parser     import ParserJSON
from .pickle_parser   import ParserPickle
from .category_parser import CategoryParser
from .config_parser   import ConfigParser




#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

import sys
import traceback

#========================================================================

class Logger(object):
	
	# DEBUG, INFO           --> stdout
	# WARNING, ERROR, FATAL --> stderr

	VERBOSITY_LEVELS = {-1: '',
						 0: ['INFO', 'FATAL'],
						 1: ['INFO', 'ERROR', 'FATAL'],
						 2: ['INFO', 'WARNING', 'ERROR', 'FATAL'],
						 3: ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'FATAL']}

	WRITER = {'DEBUG': sys.stdout, 'INFO': sys.stdout,
			  'WARNING': sys.stderr, 'ERROR': sys.stderr, 'FATAL': sys.stderr}

	# more colors and styles:
	# https://stackoverflow.com/questions/2048509/how-to-echo-with-different-colors-in-the-windows-command-line
	# https://joshtronic.com/2013/09/02/how-to-use-colors-in-command-line-output/

	GREY       = '0;37'
	WHITE      = '1;37'
	YELLOW     = '1;33',
	LIGHT_RED  = '1;31',
	RED        = '0;31'

	COLORS = {'DEBUG': WHITE, 'INFO': GREY, 'WARNING': YELLOW, 'ERROR': LIGHT_RED,'FATAL': RED}	

	def __init__(self, template, verbosity = 0):
		self.template         = template
		if isinstance(verbosity, dict):
			verbosity = verbosity['default']
		self.verbosity        = verbosity
		self.verbosity_levels = self.VERBOSITY_LEVELS[self.verbosity]


	def update_verbosity(self, verbosity = 0):
		if isinstance(verbosity, dict):
			verbosity = verbosity['default']
		self.verbosity        = verbosity
		self.verbosity_levels = self.VERBOSITY_LEVELS[self.verbosity]


	def log(self, message, message_type):
		
		# check if we need to log the message
		if message_type in self.verbosity_levels:
			color = self.COLORS[message_type]
			error_message = None
			if message_type in ['WARNING', 'ERROR', 'FATAL']:				
				error_message = traceback.format_exc()
				if not 'NoneType: None'in error_message:
					self.WRITER[message_type].write(error_message)
			uncolored_message = '[%s] %s ... %s ...\n' % (message_type, self.template, message)
			message           = "\x1b[%sm" % (color) + uncolored_message + "\x1b[0m"
			self.WRITER[message_type].write(message)
			return error_message, uncolored_message








#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

import os, sys
import copy
import time
import uuid
import pickle
import subprocess

import numpy      as np 
import tensorflow as tf

from gryffin.utilities            import Logger
from gryffin.utilities.decorators import thread

#========================================================================

class DescriptorGenerator(Logger):

	eta       = 1e-3
	max_iter  = 10**3
	def __init__(self, config):

		self.config        = config
		self.is_generating = False
		self.exec_name = '%s/descriptor_generator/generation_process.py' % self.config.get('home')

		# define registers
		self.auto_gen_descs     = {}
		self.comp_corr_coeffs   = {}
		self.gen_descs_cov      = {}
		self.min_corrs          = {}	
		self.reduced_gen_descs  = {}
		self.weights            = {}
		self.sufficient_indices = {}


	@thread
	def single_generate(self, descs, objs, feature_index, result_dict = None):

		# collect all relevant properties
		sim_dict = {}
		for prop in dir(self):
			if callable(getattr(self, prop)) or prop.startswith(('__', 'W', 'config')): continue
			sim_dict[prop] = getattr(self, prop)

		sim_dict['num_samples'] = descs.shape[0]
		sim_dict['num_descs']   = descs.shape[1]

		sim_dict['descs']       = descs
		sim_dict['objs']        = objs
		sim_dict['grid_descs']  = self.config.feature_descriptors[feature_index]
	
		identifier  = str(uuid.uuid4())[:8]
		config_name = '%s/descriptor_generation_%d_%s.pkl' % (self.config.get('scratch_dir'), feature_index, identifier)
		with open(config_name, 'wb') as content:
			pickle.dump(sim_dict, content)
		
#		FNULL = open(os.devnull, 'w')
#		subprocess.call('%s %s' % (self.exec_name, config_name), shell = True, stdout = FNULL, stderr = subprocess.STDOUT)
		subprocess.call('python %s %s' % (self.exec_name, config_name), shell = True)
		print('SUBMITTED DESC GENERATION')		
		results_name = '%s/completed_descriptor_generation_%d_%s.pkl' % (self.config.get('scratch_dir'), feature_index, identifier)

		# wait for results to be written
		while not os.path.isfile(results_name):
			time.sleep(0.05)
		current_size = 0
		while current_size != os.path.getsize(results_name):
			current_size = os.path.getsize(results_name)
			time.sleep(0.05)

		time.sleep(0.2)
		try:		
			with open(results_name, 'rb') as content:
				results = pickle.load(content)
		except EOFError:
			time.sleep(2)
			with open(results_name, 'rb') as content:
				results = pickle.load(content)
		
		self.min_corrs[feature_index]          = results['min_corrs']
		self.auto_gen_descs[feature_index]     = results['auto_gen_descs']
		self.comp_corr_coeffs[feature_index]   = results['comp_corr_coeffs']
		self.gen_descs_cov[feature_index]      = results['gen_descs_cov']
		self.reduced_gen_descs[feature_index]  = results['reduced_gen_descs']
		self.weights[feature_index]            = results['weights']	
		self.sufficient_indices[feature_index] = results['sufficient_indices']

#		print("WEIGHTS", feature_index, self.weights[feature_index])
#		print("REDUCED_DESCS", feature_index, results['reduced_gen_descs'].shape)

		result_dict[feature_index] = results['reduced_gen_descs']

		os.remove(config_name)
		os.remove(results_name)



	@thread
	def generate(self, obs_params, obs_objs):

		import time 
		start = time.time()

		self.is_generating  = True
		result_dict         = {}
		feature_types       = self.config.feature_types
		feature_descriptors = self.config.feature_descriptors

#		print('FEATURE DESCRIPTORS', feature_descriptors, np.array(feature_descriptors[0]).shape)
#		print('*'*10)
#		for element in feature_descriptors:
#			print(element)
#		print('*'*10)

		for feature_index, feature_options in enumerate(self.config.feature_options):

			if feature_types[feature_index] == 'continuous': 
				self.weights[feature_index]           = None
				self.reduced_gen_descs[feature_index] = None
				result_dict[feature_index]            = None
				continue

			if feature_descriptors[feature_index] is None:
				self.weights[feature_index]           = None
				self.reduced_gen_descs[feature_index] = None
				result_dict[feature_index]            = None
				continue

			if feature_descriptors[feature_index].shape[1] == 1:
				self.weights[feature_index]           = np.array([[1.]])
				self.reduced_gen_descs[feature_index] = feature_descriptors[feature_index]
				result_dict[feature_index]            = feature_descriptors[feature_index]
				continue

			sampled_params      = obs_params[:, feature_index].astype(np.int32)
			sampled_descriptors = feature_descriptors[feature_index][sampled_params]
			sampled_objs        = np.reshape(obs_objs, (len(obs_objs), 1)) 

			self.single_generate(sampled_descriptors, sampled_objs, feature_index, result_dict)

			# avoid parallel execution if not desired
			if not self.config.get('parallel'):
				if feature_types[feature_index] == 'continuous': continue
				while not feature_index in result_dict:
					time.sleep(0.1)

		for feature_index in range(len(self.config.feature_options)):
			if feature_types[feature_index] == 'continuous': continue
			while not feature_index in result_dict:
				time.sleep(0.1)

		gen_feature_descriptors = [result_dict[feature_index] for feature_index in range(len(result_dict.keys()))]
		self.gen_feature_descriptors = gen_feature_descriptors
		self.is_generating = False

		end = time.time()
		self.desc_gen_time = end - start


	def get_descriptors(self):
		while self.is_generating:
			time.sleep(0.1)

		if hasattr(self, 'gen_feature_descriptors'):
			print('[TIME:  ', self.desc_gen_time, '  (descriptor generation)')
			return self.gen_feature_descriptors
		else:
			return self.config.feature_descriptors


	def get_summary(self):
		
		summary = {}
		feature_types = self.config.feature_types
		if not hasattr(self, 'gen_feature_descriptors'):
			for feature_index in range(len(self.config.feature_options)):
				contribs = {}
				if feature_types[feature_index] == 'continuous': continue
				feature_descriptors = self.config.feature_descriptors[feature_index]
				if feature_descriptors is None: continue
				for desc_index in range(feature_descriptors.shape[1]):
					desc_summary_dict = {}
					desc_summary_dict['relevant_given_descriptors']     = np.arange(len(feature_descriptors[:, desc_index]))
					desc_summary_dict['given_descriptor_contributions'] = np.ones(len(feature_descriptors[:, desc_index]))
					contribs['descriptor_%d' % desc_index] = copy.deepcopy(desc_summary_dict)
				summary['feature_%d' % feature_index] = copy.deepcopy(contribs)
			return summary


		for feature_index in range(len(self.config.feature_options)):
	
			if feature_types[feature_index] == 'continuous': continue
	
			weights            = self.weights[feature_index]
			reduced_gen_descs  = self.reduced_gen_descs[feature_index]
			sufficient_indices = self.sufficient_indices[feature_index]

			if weights is None: continue
			if len(sufficient_indices) == 0: continue

			# normalize weights
			normed_weights = np.empty(weights.shape)
			for index, weight_elements in enumerate(weights):
				normed_weights[index] = weight_elements / np.sum(np.abs(weight_elements))

			# identify contributing indices
			contribs = {}
#			for new_desc_index in range(reduced_gen_descs.shape[1]):
			for new_desc_index in sufficient_indices:
				desc_summary_dict = {}
				relevant_weights  = normed_weights[new_desc_index]

				sorting_indices = np.argsort(np.abs(relevant_weights))
				cumulative_sum  = np.cumsum(np.abs(relevant_weights[sorting_indices]))
				include_indices = np.where(cumulative_sum > 0.1)[0]

				relevant_given_descriptors = sorting_indices[include_indices]
				desc_summary_dict['relevant_given_descriptors']     = relevant_given_descriptors
				desc_summary_dict['given_descriptor_contributions'] = weights[new_desc_index]
				contribs['descriptor_%d' % new_desc_index] = copy.deepcopy(desc_summary_dict)
			summary['feature_%d' % feature_index] = copy.deepcopy(contribs)

		return summary
			 


#========================================================================




#!/usr/bin/env python
  
__author__ = 'Florian Hase'

#=======================================================================

import numpy  as np 
import pandas as pd
from datetime import datetime

from utilities import Logger

#=======================================================================

class Writer(object):
	def __init__(self, file_name):
		self.file_name = file_name

	def save(self):
		pass

class CsvWriter(Writer):
	
	def __init__(self, file_name):
		Writer.__init__(self, file_name)

	def __call__(self, pd_frame):
		pd_frame.to_csv(self.file_name)


class ExcelWriter(Writer):

	def __init__(self, file_name):
		Writer.__init__(self, file_name)
		self.writer = pd.ExcelWriter(self.file_name)

	def __call__(self, pd_frame):
		pd_frame.to_excel(self.writer, 'Sheet1')

	def save(self):
		self.writer.save()

#=======================================================================


class DB_Writer(Logger):

	def __init__(self, config):
		self.config = config
		Logger.__init__(self, 'DB_Writer', self.config.get('verbosity'))


	def create_writer(self, file_name, out_format):
		if out_format == 'xlsx':
			self.writer = ExcelWriter(file_name)
		elif out_format == 'csv':
			self.writer = CsvWriter(file_name)


	def write(self, db_content, outfile, out_format):

		# create the writer
		self.create_writer(outfile, out_format)

		# convert output dict and save via pandas routine
		dataframe = pd.DataFrame.from_dict(db_content)
		self.writer(dataframe)
		self.writer.save()

	





#!/usr/bin/env python

__author__ = 'Florian Hase'

#=======================================================================

import sys

from utilities import GryffinModuleError

#=======================================================================

try:
	import sqlalchemy as sql
except ModuleNotFoundError:
	_, error_message, _ = sys.exc_info()
	extension = '\n\tTry installing the sqlalchemy package or use a different database framework instead.'
	GryffinModuleError(str(error_message) + extension)

#=======================================================================

from .sqlite_operations import AddEntry, FetchEntries, UpdateEntries
from .sqlite_database   import SqliteDatabase




#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

from .db_cache         import DB_Cache
from .db_werkzeug      import DB_Werkzeug
from .database_handler import DatabaseHandler




#!/usr/bin/env python

__author__ = 'Florian Hase'

#=======================================================================

from . import DB_Cache
from gryffin.utilities       import Logger
from gryffin.utilities       import GryffinUnknownSettingsError

#=======================================================================

class DB_Werkzeug(Logger):


	def __init__(self, config, db_attributes, verbosity = 0):
		Logger.__init__(self, 'DB_Werkzeug', verbosity = verbosity)
		self.config   = config      
		self.db_attrs = db_attributes


	def create_database(self):
		if self.config.get_db('format') == 'sqlite':
			from .sqlite_interface import SqliteDatabase
			self.database = SqliteDatabase(self.config.get_db('path'), self.db_attrs, 'db', verbosity = self.config.get('verbosity'))
		else:
			GryffinUnknownSettingsError('did not understand database format: "%s".\n\tChoose from ["none", "sqlite"]' % self.config.get_db('format')) 


	def create_cache(self):
		all_entries = self._get({})
		self.cache  = DB_Cache(self.db_attrs, all_entries, verbosity = self.config.get('verbosity'))


	def _get(self, condition):
		entries = self.db_fetch_all(condition)
		return entries


	#=== CHECK EXISTANCE ===#
	
	def db_check_existance(self, condition):
		if hasattr(self, 'cache'):
			return self._check_existance_cache(condition)
		return self._check_existance_database(condition)


	def _check_existance_cache(self, condition):
		for attr, value in condition.items():
			if not value in self.cache[attr]:
				return False
		return True


	def _check_existance_database(self, condition):
		entries = self._get(condition)
		return len(entries) > 0


	#=== ADD TO DB ===#
	
	def db_add(self, info_dict):
		if hasattr(self, 'cache'):
			self._add_cache(info_dict)
		self._add_database(info_dict)

	def _add_cache(self, info_dict):
		self.cache.add(info_dict)

	def _add_database(self, info_dict):
		try:
			self.database.add(info_dict)
		except AttributeError:
			info_dict_str = ''
			for key, item in info_dict.items():
				info_dict_str = '%s:\t%s\n' % (str(key), str(item))
			self.log('could not add to database:\n%s' % (info_dict_str), 'ERROR')


	#=== FETCH FROM DB ===#

	def db_fetch_all(self, condition_dict = {}):

		if hasattr(self, 'cache'):
			entries = self._fetch_all_cache(condition_dict)
		else:
			entries = self._fetch_all_database(condition_dict)
		return entries


	def _fetch_all_cache(self, condition_dict):
		return self.cache.fetch_all(condition_dict)

	def _fetch_all_database(self, condition_dict):
		try:
			return self.database.fetch_all(condition_dict)
		except OSError:
			condition_dict_str = ''
			for key, item in condition_dict.items():
				condition_dict_str = '%s:\t%s\n' % (str(key), str(item))
			self.log('could not fetch from database:\n%s' % (condition_dict_str), 'ERROR')

	
	#=== UPDATE DB ===#

	def db_update_all(self, condition_dict, update_dict):
		if hasattr(self, 'cache'):
			self._update_all_cache(condition_dict, update_dict)
		self._update_all_database(condition_dict, update_dict)

	def _update_all_cache(self, condition_dict, update_dict):
		self.cache.update_all(condition_dict, update_dict)

	def _update_all_database(self, condition_dict, update_dict):
		try:
			self.database.update_all(condition_dict, update_dict)
		except OSError:
			condition_dict_str = ''
			for key, item in condition_dict.items():
				condition_dict_str = '%s:\t%s\n' % (str(key), str(item))
			self.log('could not find entry with conditions in database:\n%s' % (condition_dict_str), 'ERROR')





















#!/usr/bin/env python 
  
__author__ = 'Florian Hase'

#========================================================================

import numpy as np 

from distutils.core      import setup
from distutils.extension import Extension
from Cython.Build        import cythonize

ext_modules = [
	Extension(
			'kernel_evaluations', ['BayesianNetwork/kernel_evaluations.pyx'],
			 include_dirs       = [np.get_include(), '.'],
			 extra_compile_args = ['-fopenmp'], extra_link_args = ['-fopenmp'],
		),
	Extension(
			'kernel_prob_reshaper', ['BayesianNetwork/kernel_prob_reshaping.pyx'],
			 include_dirs       = [np.get_include(), '.'],
			 extra_compile_args = ['-fopenmp'], extra_link_args = ['-fopenmp'],
		),
]

setup(
	name = 'Phoenics',
	ext_modules  = cythonize(ext_modules)
)



#!/usr/bin/env python 

# import versioneer
# import numpy as np

from setuptools import setup, find_packages
from distutils.extension import Extension
from distutils.command.build import build as build_orig

#===============================================================================

def readme():
	with open('README.md', 'r') as content:
		return content.read()

#===============================================================================

ext_modules = [
	Extension('gryffin.bayesian_network.kernel_evaluations',
		['src/gryffin/bayesian_network/kernel_evaluations.c']),
	Extension('gryffin.bayesian_network.kernel_prob_reshaping',
		['src/gryffin/bayesian_network/kernel_prob_reshaping.c']),]

# Preinstall numpy
from setuptools import dist
dist.Distribution().fetch_build_eggs(['numpy>=1.10'])
import numpy as np

def requirements():
	with open('requirements.txt', 'r') as content:
		return content.readlines()

#===============================================================================

setup(name='gryffin',
	#version=versioneer.get_version(),
	version='0.1.0',
    # cmdclass=versioneer.get_cmdclass(),
	description='Bayesian optimization for categorical variables', 
	long_description=readme(),
	long_description_content_type='text/markdown',
	classifiers=[
		'Intended Audience :: Science/Research',
		'Operating System :: Unix', 
		'Programming Language :: Python',
		'Topic :: Scientific/Engineering', 
	],
	url='https://github.com/aspuru-guzik-group/gryffin',
	author='Florian Hase',
	packages=find_packages('./src/'),
	package_dir={'': 'src'},
	zip_safe=False,
	ext_modules=ext_modules,
	tests_require=['pytest'],
	include_dirs=np.get_include(),
	install_requires=requirements(),
	python_requires='>=3.6',
)



#!/usr/bin/env python

__author__ = 'Florian Hase'

#=======================================================================

import time 
import sqlalchemy as sql

#=======================================================================

class AddEntry(object):

	def __init__(self, database, table, entry):
		self.db    = database
		self.table = table 
		self.entry = entry

	def execute(self):
		with self.db.connect() as conn:
			conn.execute(self.table.insert(), self.entry)
			conn.close()

#=======================================================================

class FetchEntries(object):
	
	def __init__(self, database, table, selection):
		self.db              = database
		self.table           = table
		self.selection       = selection
		self.entries         = None
		self.executed        = False
		self.entries_fetched = False

	def execute(self):
		with self.db.connect() as conn:
			selected  = conn.execute(self.selection)
			entries   = selected.fetchall()
			conn.close()
		self.entries  = entries
		self.executed = True

	def get_entries(self):
		iteration_index = 0
		while not self.executed:
			time.sleep(0.02)
		self.entries_fetched = True
		return self.entries

#=======================================================================

class UpdateEntries(object):

	def __init__(self, database, table, updates):
		self.db      = database
		self.table   = table
		self.updates = updates


	def execute(self):
		with self.db.connect() as conn:
			updated = conn.execute(self.updates)
			conn.close()













#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

import cython

from cython.parallel import prange

import  numpy as np 
cimport numpy as np 

from libc.math cimport exp, sqrt

#========================================================================

cdef class KernelReshaper:

	cdef int num_samples, num_obs, num_kernels, num_descriptors
	cdef np.ndarray np_recomputed_probs
	cdef np.ndarray np_all_distances

	def __init__(self):

		pass


	@cython.cdivision(True)
	@cython.boundscheck(False)
	cdef double [:, :, :] _reshape_probs(self, double [:, :, :] cat_probs, double [:, :] descriptors):

		cdef double [:, :, :] recomputed_probs = self.np_recomputed_probs
		cdef double [:, :, :] all_distances    = self.np_all_distances

		cdef double ds2, dyi, sum_distances
		cdef double averaged_descriptor

		cdef int sample_index, obs_index, target_cat_index, desc_index, kernel_index

		for sample_index in prange(self.num_samples, nogil = True):

			for obs_index in range(self.num_obs):

				for desc_index in range(self.num_descriptors):
	
					averaged_descriptor = 0.
					for kernel_index in range(self.num_kernels):
						averaged_descriptor = cat_probs[sample_index, obs_index, kernel_index] * descriptors[kernel_index, desc_index] + averaged_descriptor	


				for target_cat_index in range(self.num_kernels):

					ds2 = 0.
					for desc_index in range(self.num_descriptors):
					
						dyi = self.num_kernels * (descriptors[target_cat_index, desc_index] - averaged_descriptor)
						ds2 = ds2 + dyi * dyi

					all_distances[sample_index, obs_index, target_cat_index] = sqrt(ds2 / self.num_descriptors)

				# got all distances, compute probs from distances
				sum_distances = 0.
				for kernel_index in range(self.num_kernels):
					sum_distances = sum_distances + exp( - all_distances[sample_index, obs_index, kernel_index])
				
				for kernel_index in range(self.num_kernels):
					recomputed_probs[sample_index, obs_index, kernel_index] = exp( - all_distances[sample_index, obs_index, kernel_index]) / sum_distances

		return recomputed_probs



	cpdef reshape_probs(self, np.ndarray cat_probs, np.ndarray descriptors):

		self.num_samples     = cat_probs.shape[0]
		self.num_obs         = cat_probs.shape[1]
		self.num_kernels     = cat_probs.shape[2]
		self.num_descriptors = descriptors.shape[1]

		self.np_recomputed_probs = np.zeros((self.num_samples, self.num_obs, self.num_kernels))
		self.np_all_distances    = np.zeros((self.num_samples, self.num_obs, self.num_kernels))

		cdef double [:, :, :] cat_probs_memview   = cat_probs
		cdef double [:, :]    descriptors_memview = descriptors

		reshaped_probs = self._reshape_probs(cat_probs_memview, descriptors_memview)		
		return np.array(reshaped_probs)




#!/usr/bin/env python 

__author__ = 'Florian Hase'

#========================================================================

from gryffin.utilities import Logger

#========================================================================

class DB_Cache(Logger):

	def __init__(self, attributes, entries = [], verbosity = 0):
		Logger.__init__(self, 'DB_Cache', verbosity = verbosity)
		self.attributes = attributes		

		self.cache     = {attr: [] for attr in self.attributes}
		self.num_items = 0
		for entry in entries:
			self.add(entry)


	def __getitem__(self, item):
		try:
			return self.cache[item]
		except KeyError:
			return []



	def add(self, info_dict):
		for attr in self.attributes:
			if attr in info_dict:
				self.cache[attr].append(info_dict[attr])
			else:	
				self.cache[attr].append(None)
		self.num_items += 1


	def fetch_all(self, condition_dict):
		results = []
		for element_index in range(self.num_items):
			for key, value in condition_dict.items():
				if value != self.cache[key][element_index]:
					break
			else:
				result = {attr: self.cache[attr][element_index] for attr in self.attributes}
				results.append(result)
		return results


	def update_all(self, condition_dict, update_dict):
		for element_index in range(self.num_items):
			for key, value in condition_dict.items():
				if value != self.cache[key][element_index]:
					break
			else:
				for key, value in update_dict.items():
					self.cache[key][element_index] = value





#!/usr/bin/env python 
  
__author__ = 'Florian Hase'

#========================================================================

import numpy as np 
import multiprocessing
from multiprocessing import Process, Manager

from . import ParameterOptimizer
from gryffin.random_sampler import RandomSampler
from gryffin.utilities      import Logger

#========================================================================

class Acquisition(Logger):

	def __init__(self, config):
	
		self.config = config
		Logger.__init__(self, 'Acquisition', self.config.get('verbosity'))
		self.random_sampler   = RandomSampler(self.config.general, self.config.parameters)
		self.total_num_vars   = len(self.config.feature_names)
		self.local_optimizers = None		
		self.num_cpus         = multiprocessing.cpu_count()


	def _propose_randomly(self, best_params, num_samples, dominant_samples = None):
		# get uniform samples
		if dominant_samples is None:
			uniform_samples = self.random_sampler.draw(num = self.total_num_vars * num_samples)	
			perturb_samples = self.random_sampler.perturb(best_params, num = self.total_num_vars * num_samples)
			samples         = np.concatenate([uniform_samples, perturb_samples])
		else:
			dominant_features = self.config.feature_process_constrained
			for batch_sample in dominant_samples:
				uniform_samples = self.random_sampler.draw(num = self.total_num_vars * num_samples // len(dominant_samples))
				perturb_samples = self.random_sampler.perturb(best_params, num = self.total_num_vars * num_samples)
				samples         = np.concatenate([uniform_samples, perturb_samples])
			samples[:, dominant_features] = batch_sample[dominant_features]
		return samples


	def _proposal_optimization_thread(self, proposals, kernel_contribution, batch_index, return_index, return_dict = None, dominant_samples = None):
		self.log('starting process for %d' % batch_index, 'INFO')

		def kernel(x):
			num, inv_den = kernel_contribution(x)
			return (num + self.sampling_param_values[batch_index]) * inv_den 
	
		if dominant_samples is not None:
			ignore = self.config.feature_process_constrained
		else:
			ignore = np.array([False for _ in range(len(self.config.feature_process_constrained))])

		local_optimizer = self.local_optimizers[batch_index]
		local_optimizer.set_func(kernel, ignores = ignore)	

		optimized = []
		for sample_index, sample in enumerate(proposals):
			opt = local_optimizer.optimize(kernel, sample, max_iter = 10)
			optimized.append(opt)
		optimized = np.array(optimized)

		if return_dict.__class__.__name__ == 'DictProxy':
			return_dict[return_index] = optimized
		else:
			return optimized	


	def _optimize_proposals(self, random_proposals, kernel_contribution, dominant_samples = None):

		if self.config.get('parallel'):
			result_dict = Manager().dict()

			# get the number of splits
			num_splits = self.num_cpus // len(self.sampling_param_values) + 1
			split_size = len(random_proposals) // num_splits 

			processes   = []
			for batch_index in range(len(self.sampling_param_values)):
				for split_index in range(num_splits):

					split_start  = split_size * split_index
					split_end    = split_size * (split_index + 1)
					return_index = num_splits * batch_index + split_index
					process = Process(target = self._proposal_optimization_thread, args = (random_proposals[split_start : split_end], kernel_contribution, batch_index, return_index, result_dict, dominant_samples))
					processes.append(process)
					process.start()

			for process_index, process in enumerate(processes):
				process.join()	

		else:
			num_splits  = 1
			result_dict = {}
			for batch_index in range(len(self.sampling_param_values)):
				return_index = batch_index
				result_dict[batch_index] = self._proposal_optimization_thread(random_proposals, kernel_contribution, batch_index, return_index, dominant_samples = dominant_samples)

		# collect optimized samples
		samples = []
		for batch_index in range(len(self.sampling_param_values)):
			batch_samples = []
			for split_index in range(num_splits):
				return_index = num_splits * batch_index + split_index
				batch_samples.append(result_dict[return_index])
			samples.append(np.concatenate(batch_samples))
		samples = np.array(samples)
		return np.array(samples)




	def propose(self, best_params, kernel_contribution, sampling_param_values, 
				num_samples = 200, 
				parallel = 'True',
				dominant_samples  = None,
				dominant_strategy = None,
			):

		self.local_optimizers = [ParameterOptimizer(self.config) for _ in range(len(sampling_param_values))]
		assert len(self.local_optimizers) == len(sampling_param_values)

		self.sampling_param_values = sampling_param_values
		random_proposals = self._propose_randomly(
				best_params, num_samples, dominant_samples = dominant_samples,
			)


		import time
		start = time.time()
		optimized_proposals = self._optimize_proposals(
				random_proposals, kernel_contribution, dominant_samples = dominant_samples,
			)
		end   = time.time()
		#print('[TIME]:  ', end - start, '  (optimizing proposals)')
		self.log('[TIME]:  ' + str(end - start) + '  (optimizing proposals)', 'INFO')


		extended_proposals = np.array([random_proposals for _ in range(len(sampling_param_values))])
		combined_proposals = np.concatenate((extended_proposals, optimized_proposals), axis = 1)

		return combined_proposals

